import copy
import inspect
import io
import random
import warnings
import numpy as np
import psutil
import tempfile312
from numba import UnsupportedError

from matplotlib import pyplot as plt

from RuleTree.stumps.classification.DecisionTreeStumpClassifier import DecisionTreeStumpClassifier
from RuleTree.utils.define import DATA_TYPE_TABULAR
from RuleTree.utils.shapelet_transform.TabularShapelets import TabularShapelets


class PartialPivotTreeStumpClassifier(DecisionTreeStumpClassifier):
    def __init__(self, n_shapelets=psutil.cpu_count(logical=False)*2,
                 n_shapelets_for_selection=500,  #int, inf, or 'stratified'
                 n_ts_for_selection=100,  #int, inf
                 n_features_strategy=2,
                 selection='mi_clf',  #random, mi_clf, mi_reg, cluster
                 distance='euclidean',
                 mi_n_neighbors=100,
                 random_state=42, n_jobs=1,
                 **kwargs):
        self.n_shapelets = n_shapelets
        self.n_shapelets_for_selection = n_shapelets_for_selection
        self.n_ts_for_selection = n_ts_for_selection
        self.n_features_strategy = n_features_strategy
        self.selection = selection
        self.distance = distance
        self.mi_n_neighbors = mi_n_neighbors
        self.random_state = random_state
        self.n_jobs = n_jobs

        if "max_depth" in kwargs and kwargs["max_depth"] > 1:
            warnings.warn("max_depth must be 1")

        kwargs["max_depth"] = 1

        if selection not in ['random', 'mi_clf', 'cluster']:
            raise ValueError("'selection' must be 'random', 'mi_clf' or 'cluster'")

        super().__init__(**kwargs)

        self.kwargs |= {
            "n_shapelets": n_shapelets,
            "n_shapelets_for_selection": n_shapelets_for_selection,
            "n_ts_for_selection": n_ts_for_selection,
            "n_features_strategy": n_features_strategy,
            "selection": selection,
            "distance": distance,
            "mi_n_neighbors": mi_n_neighbors,
            "random_state": random_state,
            "n_jobs": n_jobs,
        }

    def fit(self, X, y, idx=None, context=None, sample_weight=None, check_input=True):
        if idx is None:
            idx = slice(None)
        X = X[idx]
        y = y[idx]

        self.y_lims = [X.min(), X.max()]

        random.seed(self.random_state)
        if sample_weight is not None:
            raise UnsupportedError(f"sample_weight is not supported for {self.__class__.__name__}")

        self.st = TabularShapelets(n_shapelets=self.n_shapelets,
                                   n_shapelets_for_selection=self.n_shapelets_for_selection,
                                   n_ts_for_selection=self.n_ts_for_selection,
                                   n_features_strategy=self.n_features_strategy,
                                   selection=self.selection,
                                   distance=self.distance,
                                   mi_n_neighbors=self.mi_n_neighbors,
                                   random_state=random.randint(0, 2**32-1),
                                   n_jobs=self.n_jobs
                                   )

        return super().fit(self.st.fit_transform(X, y), y=y, sample_weight=sample_weight, check_input=check_input)

    def apply(self, X, check_input=False):
        self.y_lims = [min(self.y_lims[0], X.min()), min(self.y_lims[1], X.max())]

        return super().apply(self.st.transform(X), check_input=check_input)

    def supports(self, data_type):
        return data_type in [DATA_TYPE_TABULAR]

    def get_rule(self, columns_names=None, scaler=None, float_precision: int | None = 3):
        rule = {
            "feature_idx": self.feature_original[0],
            "threshold": self.threshold_original[0],
            "is_categorical": self.is_categorical,
            "samples": self.n_node_samples[0]
        }

        rule["feature_name"] = f"PartialPivot_{rule['feature_idx']}"

        if scaler is not None:
            raise UnsupportedError(f"Scaler not supported for {self.__class__.__name__}")

        comparison = "<="
        not_comparison = ">"
        rounded_value = str(rule["threshold"]) if float_precision is None else round(rule["threshold"], float_precision)

        shape = self.st.shapelets[self.feature_original[0], 0]

        with tempfile312.NamedTemporaryFile(delete_on_close=False,
                                            delete=False,
                                            suffix=".png",
                                            mode="wb") as temp_file:
            plt.figure(figsize=(2, 1))
            plt.plot([i for i in range(shape.shape[0])], shape)
            plt.subplots_adjust(left=0, right=1, top=1, bottom=0)
            plt.ylim(*self.y_lims)
            plt.xlim(0, shape.shape[0])
            plt.gca().tick_params(axis='both', which='both', length=2, labelsize=6)
            plt.gca().spines['right'].set_color('none')
            plt.gca().spines['top'].set_color('none')
            #plt.gca().spines['bottom'].set_position('zero')
            plt.savefig(temp_file, format="png", dpi=300, bbox_inches='tight', pad_inches=0)
            plt.close()

        rule["textual_rule"] = f"{self.distance}(PP, shp) {comparison} {rounded_value}\t{rule['samples']}"
        rule["blob_rule"] = f"{self.distance}(PP, shp) {comparison} {rounded_value}"
        rule["graphviz_rule"] = {
            "image": f'{temp_file.name}',
            "imagescale": "true",
            "imagepos": "bc",
            "label": f"{self.distance}(PP, shp) \u2264 {rounded_value}",
            "labelloc": "t",
            "fixedsize": "true",
            "width": "2",
            "height": "1.33",
            "shape": "none",
            "fontsize": "10",
        }

        rule["not_textual_rule"] = f"{self.distance}(PP, shp) {not_comparison} {rounded_value}"
        rule["not_blob_rule"] = f"{self.distance}(PP, shp) {not_comparison} {rounded_value}"
        rule["not_graphviz_rule"] = {
            "image": f'{temp_file.name}',
            "imagescale": "true",
            "label": f"{self.distance}(PP, shp) {not_comparison} {rounded_value}",
            "imagepos": "bc",
            "labelloc": "t",
            "fixedsize": "true",
            "width": "2",
            "height": "1.33",
            "shape": "none",
            "fontsize": "10",
        }

        return rule

    def node_to_dict(self):
        rule = super().node_to_dict() | {
            'stump_type': self.__class__.__module__,
            "feature_idx": self.feature_original[0],
            "threshold": self.threshold_original[0],
            "is_categorical": self.is_categorical,
            "samples": self.n_node_samples[0]
        }

        rule["feature_name"] = f"PartialPivot_{rule['feature_idx']}"

        comparison = "<="
        not_comparison = ">"
        rounded_value = rule["threshold"]

        rule["textual_rule"] = f"{self.distance}(PP, shp) {comparison} {rounded_value}\t{rule['samples']}"
        rule["blob_rule"] = f"{self.distance}(PP, shp) {comparison} {rounded_value}"
        rule["graphviz_rule"] = {
            "image": f'None',
            "imagescale": "true",
            "imagepos": "bc",
            "label": f"{self.distance}(PP, shp) {comparison} {rounded_value}",
            "labelloc": "t",
            "fixedsize": "true",
            "width": "2",
            "height": "1.33",
            "shape": "none",
            "fontsize": "10",
        }

        rule["not_textual_rule"] = f"{self.distance}(PP, shp) {not_comparison} {rounded_value}"
        rule["not_blob_rule"] = f"{self.distance}(PP, shp) {not_comparison} {rounded_value}"
        rule["not_graphviz_rule"] = {
            "image": f'{None}',
            "imagescale": "true",
            "label": f"{self.distance}(PP, shp) {not_comparison} {rounded_value}",
            "imagepos": "bc",
            "labelloc": "t",
            "fixedsize": "true",
            "width": "2",
            "height": "1.33",
            "shape": "none",
            "fontsize": "10",
        }

        # shapelet transform stuff
        rule["shapelets"] = self.st.shapelets.tolist()
        rule["n_shapelets"] = self.st.n_shapelets
        rule["n_shapelets_for_selection"] = self.st.n_shapelets_for_selection
        rule["n_ts_for_selection_per_class"] = self.st.n_ts_for_selection
        rule["min_n_features"] = self.st.min_n_features
        rule["max_n_features"] = self.st.max_n_features
        rule["selection"] = self.st.selection
        rule["distance"] = self.st.distance
        rule["mi_n_neighbors"] = self.st.mi_n_neighbors
        rule["random_state"] = self.st.random_state
        rule["n_jobs"] = self.st.n_jobs
        rule["y_lims"] = self.y_lims

        return rule

    @classmethod
    def dict_to_node(cls, node_dict, X=None):
        self = cls(
            n_shapelets=node_dict["n_shapelets"],
            n_shapelets_for_selection=node_dict["n_shapelets_for_selection"],
            n_ts_for_selection=node_dict["n_ts_for_selection_per_class"],
            sliding_window=node_dict["sliding_window"],
            selection=node_dict["selection"],
            distance=node_dict["distance"],
            mi_n_neighbors=node_dict["mi_n_neighbors"],
            random_state=node_dict["random_state"],
            n_jobs=node_dict["n_jobs"]
        )

        self.st = TabularShapelets(
            n_shapelets=node_dict["n_shapelets"],
            n_shapelets_for_selection=node_dict["n_shapelets_for_selection"],
            n_ts_for_selection=node_dict["n_ts_for_selection_per_class"],
            min_n_features=node_dict["min_n_features"],
            max_n_features=node_dict["max_n_features"],
            selection=node_dict["selection"],
            distance=node_dict["distance"],
            mi_n_neighbors=node_dict["mi_n_neighbors"],
            random_state=node_dict["random_state"],
            n_jobs=node_dict["n_jobs"]
        )

        self.st.shapelets = np.array(node_dict["shapelets"])

        self.feature_original = np.zeros(3, dtype=int)
        self.threshold_original = np.zeros(3)
        self.n_node_samples = np.zeros(3, dtype=int)

        self.feature_original[0] = node_dict["feature_idx"]
        self.threshold_original[0] = node_dict["threshold"]
        self.n_node_samples[0] = node_dict["samples"]
        self.is_categorical = node_dict["is_categorical"]

        self.y_lims = node_dict["y_lims"]

        args = copy.deepcopy(node_dict["args"])
        self.is_oblique = args.pop("is_oblique")
        self.is_pivotal = args.pop("is_pivotal")
        self.unique_val_enum = args.pop("unique_val_enum")
        self.coefficients = args.pop("coefficients")
        self.kwargs = args

        return self


import numpy as np
import pandas as pd
from sklearn.metrics import mean_squared_error, mean_absolute_error, mean_poisson_deviance
from sklearn.tree import DecisionTreeRegressor

from RuleTree.base.RuleTreeBaseStump import RuleTreeBaseStump


class ObliqueDecisionTreeStumpRegressor(DecisionTreeRegressor, RuleTreeBaseStump):
    """
    Oblique Decision Tree Stump Regressor for regression tasks.
    
    This class extends DecisionTreeRegressor to create decision stumps with oblique splits,
    which allows for multivariate splits using linear combinations of features rather than
    splits on a single feature.
    
    Inherits from:
        DecisionTreeRegressor: sklearn's standard decision tree for regression
        RuleTreeBaseStump: base class for rule tree stumps
    """

    def get_rule(self, columns_names=None, scaler=None, float_precision=3):
        """
        Get a textual representation of the rule defined by this stump.
        
        Parameters:
        -----------
        columns_names : list, optional
            Names of the columns/features for creating readable rules
        scaler : object, optional
            Scaler object for transforming numeric values if the data was scaled
        float_precision : int, optional (default=3)
            Number of decimal places for rounding numeric values in the rule
            
        Returns:
        --------
        dict
            Rule representation not yet implemented for this class
            
        Raises:
        -------
        NotImplementedError
            This method is not yet implemented
        """
        raise NotImplementedError()

    def node_to_dict(self, col_names):
        """
        Convert the node to a dictionary representation.
        
        Parameters:
        -----------
        col_names : list
            List of column names
            
        Returns:
        --------
        dict
            Dictionary representation of the node
            
        Raises:
        -------
        NotImplementedError
            This method is not yet implemented
        """
        raise NotImplementedError()

    def export_graphviz(self, graph=None, columns_names=None, scaler=None, float_precision=3):
        """
        Export the stump as a graphviz visualization.
        
        Parameters:
        -----------
        graph : pydot.Graph, optional
            Existing graph object to add nodes to
        columns_names : list, optional
            Names of the columns/features
        scaler : object, optional
            Scaler object for transforming numeric values
        float_precision : int, optional (default=3)
            Number of decimal places for rounding numeric values
            
        Returns:
        --------
        pydot.Graph
            Graphviz representation of the stump
            
        Raises:
        -------
        NotImplementedError
            This method is not yet implemented
        """
        raise NotImplementedError()

    def __init__(self, **kwargs):
        """
        Initialize the Oblique Decision Tree Stump Regressor.
        
        Parameters:
        -----------
        **kwargs : dict
            Parameters to be passed to sklearn's DecisionTreeRegressor
        """
        super().__init__(**kwargs)
        self.oblique_split = None
        self.is_categorical = False
        self.kwargs = kwargs
        self.unique_val_enum = None
        self.threshold_original = None
        self.feature_original = None

        if kwargs['criterion'] == "squared_error":
            self.impurity_fun = mean_squared_error
        elif kwargs['criterion'] == "friedman_mse":
            raise Exception("not implemented")  # TODO: implement
        elif kwargs['criterion'] == "absolute_error":
            self.impurity_fun = mean_absolute_error
        elif kwargs['criterion'] == "poisson":
            self.impurity_fun = mean_poisson_deviance
        else:
            self.impurity_fun = kwargs['criterion']

    def __impurity_fun(self, **x):
        """
        Calculate impurity for a node.
        
        Parameters:
        -----------
        **x : dict
            Keyword arguments containing at least y_true
            
        Returns:
        --------
        float
            Impurity value (0 if y_true is empty)
        """
        return self.impurity_fun(**x) if len(x["y_true"]) > 0 else 0  # TODO: check

    def get_params(self, deep=True):
        """
        Get the parameters of this estimator.
        
        Parameters:
        -----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained sub-objects that are estimators.
            
        Returns:
        --------
        dict
            Parameter names mapped to their values
        """
        return self.kwargs

    def fit(self, X, y, idx=None, context=None, sample_weight=None, check_input=True):
        """
        Build an oblique decision tree stump regressor from the training data.
        
        Parameters:
        -----------
        X : array-like of shape (n_samples, n_features)
            The training input samples
        y : array-like of shape (n_samples,)
            The target values
        idx : slice or array-like, optional
            Indices of samples to use for training
        context : object, optional
            Additional context information (not used)
        sample_weight : array-like of shape (n_samples,), optional
            Sample weights
        check_input : bool, default=True
            Whether to check the input parameters
            
        Returns:
        --------
        self : object
            Fitted estimator
        """
        if idx is None:
            idx = slice(None)
        X = X[idx]
        y = y[idx]

        dtypes = pd.DataFrame(X).infer_objects().dtypes
        self.numerical = dtypes[dtypes != np.dtype('O')].index
        self.categorical = dtypes[dtypes == np.dtype('O')].index

        if len(self.numerical) > 0:
            self.oblique_split.fit(X[:, self.numerical], y, sample_weight=sample_weight, check_input=check_input)
            X_transform = self.oblique_split.transform(X[:, self.numerical])
            super().fit(X_transform, y, sample_weight=sample_weight, check_input=check_input)

            self.feature_original = [self.oblique_split.feats, -2, -2]
            self.coefficients = self.oblique_split.coeff
            self.threshold_original = self.tree_.threshold

        return self

    def apply(self, X):
        """
        Apply the decision stump to X and return the leaf indices.
        
        Parameters:
        -----------
        X : array-like of shape (n_samples, n_features)
            The input samples to apply the stump on
            
        Returns:
        --------
        X_leaves : array-like of shape (n_samples,)
            The leaf indices for each sample
        """
        X_transform = self.oblique_split.transform(X[:, self.numerical])
        return super().apply(X_transform)



import itertools

import numpy as np
from sklearn.metrics import pairwise_distances
from sklearn.tree import DecisionTreeClassifier
from inspect import signature

from RuleTree.stumps.splitters.ObliquePivotSplit import ObliquePivotSplit
from RuleTree.utils import MODEL_TYPE_CLU, MODEL_TYPE_REG, MODEL_TYPE_CLF
from RuleTree.utils.data_utils import get_info_gain


class MultipleObliquePivotSplit(ObliquePivotSplit):
    """
    MultipleObliquePivotSplit extends ObliquePivotSplit to find the best pair of pivot instances for splits.
    
    This class evaluates all possible pairs of pivot instances identified by the ObliquePivotSplit
    method and selects the pair that yields the highest information gain when used for splitting the data.
    
    Parameters
    ----------
    **kwargs : dict
        Additional parameters to pass to the parent ObliquePivotSplit class.
    
    Attributes
    ----------
    best_tup : array-like
        Best tuple of pivot instances for splitting.
    best_tup_name : array-like
        Names/indices of the best tuple of pivot instances.
    best_gain : float
        Information gain achieved by the best tuple.
    """
    def __init__(
            self,
            **kwargs
    ):
        super().__init__(**kwargs)
        self.best_tup = None
        self.best_tup_name = None
        self.best_gain = -float('inf')

    def find_best_tuple(self, X, y, distance_measure='euclidean', sample_weight=None, check_input=True):
        """
        Find the best pair of pivot instances that maximizes information gain.
        
        Evaluates all possible pairs of candidate instances and selects the pair
        that yields the highest information gain when used for splitting.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input feature matrix.
        y : array-like of shape (n_samples,)
            Target values.
        distance_measure : str or callable, default='euclidean'
            Distance measure to use.
        sample_weight : array-like, optional
            Sample weights.
        check_input : bool, default=True
            Whether to validate input.
        """
        two_tuples = list(itertools.combinations(range(0, len(self.X_candidates)), 2))
    
        for tup in two_tuples:
            # Skip pairs from the same class
            if len(set(self.y_candidates[np.array(tup)])) == 1:
                continue
                
            disc = self.get_base_model_for_tuple_finding()
           
            p1, p2 = self.X_candidates[np.array(tup)]
            name_p1, name_p2 = self.candidates_names[np.array(tup)]

            # Calculate distances to both pivot instances
            dist_to_p0 = pairwise_distances(X, p1.reshape(1, -1), metric=distance_measure).flatten()
            dist_to_p1 = pairwise_distances(X, p2.reshape(1, -1), metric=distance_measure).flatten()

            # Create binary features: Is point closer to p1 or p2?
            dist_binary = np.where(dist_to_p0 < dist_to_p1, 0, 1).reshape(-1, 1)
            disc.fit(dist_binary, y)
            gain_disc = get_info_gain(disc)

            # Update if better gain found
            if gain_disc > self.best_gain:
                self.best_gain = gain_disc
                self.best_tup = self.X_candidates[np.array(tup)]
                self.best_tup_name = self.candidates_names[np.array(tup)]

    def fit(self, X, y, distance_matrix, distance_measure, idx,
            sample_weight=None, check_input=True):
        """
        Fit the MultipleObliquePivotSplit.
        
        First calls the parent ObliquePivotSplit fit method to identify candidates,
        then finds the best pair of pivot instances for splitting.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input feature matrix.
        y : array-like of shape (n_samples,)
            Target values.
        distance_matrix : array-like of shape (n_samples, n_samples)
            Pre-computed distance matrix between instances.
        distance_measure : str or callable
            Distance measure to use.
        idx : array-like of shape (n_samples,)
            Indices of the instances.
        sample_weight : array-like, optional
            Sample weights.
        check_input : bool, default=True
            Whether to validate input.
            
        Returns
        -------
        self
            The fitted splitter.
        """
        super().fit(X, y, distance_matrix, distance_measure, idx, sample_weight=sample_weight, check_input=check_input)
        self.find_best_tuple(X, y, distance_measure=distance_measure, sample_weight=sample_weight,
                             check_input=check_input)

    def transform(self, X, distance_measure='euclidean'):
        """
        Transform input data using the best pair of pivot instances.
        
        For each input instance, determines whether it's closer to the first or
        second pivot instance in the best pair.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Input feature matrix.
        distance_measure : str or callable, default='euclidean'
            Distance measure to use.
            
        Returns
        -------
        array-like of shape (n_samples, 1)
            Binary feature indicating whether each instance is closer to the first (0)
            or second (1) pivot instance.
        """
        dist_to_p0 = pairwise_distances(X, self.best_tup[0].reshape(1, -1), metric=distance_measure).flatten()
        dist_to_p1 = pairwise_distances(X, self.best_tup[1].reshape(1, -1), metric=distance_measure).flatten()
        dist_binary = np.where(dist_to_p0 < dist_to_p1, 0, 1).reshape(-1, 1)
        return dist_binary

    def get_best_tup_names(self):
        """
        Returns the names/indices of the best pair of pivot instances.
        
        Returns
        -------
        array-like
            Names/indices of the best pair of pivot instances.
        """
        return self.best_tup_name

    def get_base_model_for_tuple_finding(self):
        """
        Returns the appropriate base model for tuple finding based on the machine learning task.
        
        Filters the kwargs to include only valid parameters for the selected model.
        
        Returns
        -------
        model : estimator
            The machine learning model to use for finding the best tuple.
            
        Raises
        ------
        NotImplementedError
            If the ml_task is regression or clustering, which are not yet implemented.
        """
        if self.ml_task == MODEL_TYPE_CLF:
            valid_params = set(signature(DecisionTreeClassifier).parameters.keys())
            filtered_kwargs = {key: value for key, value in self.kwargs.items() if key in valid_params}
            return DecisionTreeClassifier(**filtered_kwargs)
        elif self.ml_task == MODEL_TYPE_REG:
            raise NotImplementedError("Regression is not implemented yet.")
        elif self.ml_task == MODEL_TYPE_CLU:
            raise NotImplementedError("Clustering is not implemented yet.")



from RuleTree.base.RuleTreeBaseStump import RuleTreeBaseStump
from RuleTree.stumps.classification.DecisionTreeStumpClassifier import DecisionTreeStumpClassifier
from RuleTree.stumps.splitters.PivotSplit import PivotSplit
from RuleTree.utils import MODEL_TYPE_CLF
from sklearn.metrics.pairwise import pairwise_distances
import numpy as np
import copy
import warnings


class PivotTreeStumpClassifier(DecisionTreeStumpClassifier, RuleTreeBaseStump):
    """
    A classifier that uses a pivot-based splitting strategy for decision tree stumps.

    This classifier implements a pivot-based approach for decision tree stumps in classification tasks.
    It works by selecting a pivot instance from the dataset and calculating distances from all other
    instances to this pivot. These distances are then used to make splitting decisions.

    The pivot-based approach is particularly useful for working with non-standard data representations
    where traditional feature-based splits may not be applicable. It provides an alternative
    splitting mechanism based on distance metrics.

    This classifier supports both numerical and categorical features and inherits functionality
    from both DecisionTreeStumpClassifier and RuleTreeBaseStump.
    """

    def __init__(self, **kwargs):
        """
        Initialize the PivotTreeStumpClassifier.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments passed to the parent classes and the PivotSplit.
            Common parameters include criterion, max_depth, min_samples_split, etc.
        """
        super().__init__(**kwargs)
        self.pivot_split = PivotSplit(ml_task=MODEL_TYPE_CLF, **kwargs)
        self.distance_measure = None
        self.split_instance = None

    def fit(self, X, y, distance_matrix, distance_measure, idx, sample_weight=None, check_input=True):
        """
        Fit the classifier to the training data.

        This method implements the model training process by:
        1. Analyzing the features in the input data
        2. Computing pivot-based splits using the provided distance matrix
        3. Transforming the data based on distances to the pivot instance
        4. Fitting the underlying decision tree stump to the transformed data

        Parameters
        ----------
        X : numpy.ndarray
            Feature matrix of shape (n_samples, n_features).
        y : numpy.ndarray
            Target labels of shape (n_samples,).
        distance_matrix : numpy.ndarray
            Precomputed distance matrix between samples.
        distance_measure : str
            Distance metric to use (e.g., 'euclidean', 'manhattan', 'cosine').
        idx : int
            Index of the pivot instance in the dataset.
        sample_weight : numpy.ndarray, optional
            Sample weights of shape (n_samples,). Default is None.
        check_input : bool, optional
            Whether to validate input data. Default is True.

        Returns
        -------
        self : PivotTreeStumpClassifier
            Fitted classifier instance.
        """
        self.feature_analysis(X, y)
        self.num_pre_transformed = self.numerical
        self.cat_pre_transformed = self.categorical
       
        if len(self.numerical) > 0:
            self.pivot_split.fit(X[:, self.numerical], y, distance_matrix, distance_measure, idx,
                                 sample_weight=sample_weight, check_input=check_input)
            X_transform = self.pivot_split.transform(X[:, self.numerical], distance_measure)
            candidate_names = self.pivot_split.get_candidates_names()
            super().fit(X_transform, y, sample_weight=sample_weight, check_input=check_input)

            self.feature_original = [f'{candidate_names[self.tree_.feature[0]]}', -2, -2]
            self.threshold_original = self.tree_.threshold
            self.is_pivotal = True
            
            self.distance_measure = distance_measure
            self.X_split_instance = self.pivot_split.X_candidates[self.tree_.feature[0]]

        return self

    def apply(self, X):
        """
        Apply the fitted classifier to the input data.

        This method transforms the input data using the pivot-based approach
        and then applies the classifier decision rule.

        Parameters
        ----------
        X : numpy.ndarray
            Feature matrix of shape (n_samples, n_features).

        Returns
        -------
        numpy.ndarray
            Predicted class labels for each sample in X.
        """
        X_transformed = pairwise_distances(X[:, self.num_pre_transformed], 
                                           self.X_split_instance.reshape(1, -1),
                                           metric=self.distance_measure)
        
        y_pred = (np.ones(X_transformed.shape[0]) * 2)
        X_feature = X_transformed[:,  0]
        y_pred[X_feature <= self.threshold_original[0]] = 1
        
        return y_pred
        
        #return super().apply_sk(X_transformed)
        
        
    def get_rule(self, columns_names=None, scaler=None, float_precision=3):
        """
        Get the rule representation of this classifier.

        This method generates a dictionary containing rule information,
        including feature indices, thresholds, and human-readable representations
        of the decision rule.

        Parameters
        ----------
        columns_names : list, optional
            List of column names for the features. Default is None.
        scaler : object, optional
            Scaling object used to transform feature values. Default is None.
        float_precision : int, optional
            Number of decimal places for floating point values in rules. Default is 3.

        Returns
        -------
        dict
            A dictionary containing rule information with keys:
            - feature_idx: Feature index in the original feature space
            - threshold: Threshold value for the decision rule
            - coefficients: Feature coefficients
            - is_categorical: Whether the feature is categorical
            - samples: Number of samples in the node
            - feature_name: Name of the feature
            - textual_rule: Human-readable rule representation
            - blob_rule: Simplified rule representation
            - graphviz_rule: Rule representation for Graphviz visualization
            - not_textual_rule: Negated rule representation
            - not_blob_rule: Negated simplified rule representation
            - not_graphviz_rule: Negated rule representation for Graphviz visualization
        """
        rule = {
            "feature_idx": self.feature_original[0],
            "threshold": self.threshold_original[0],
            "coefficients" : self.coefficients,
            "is_categorical": self.is_categorical,
            "samples": self.n_node_samples[0]
        }

        feat_name = f"P_{rule['feature_idx']}"
        if columns_names is not None:
            #feat_names should not be useful for pivot tree
            #feat_name = columns_names[self.feature_original[0]]
            feat_name = None
        rule["feature_name"] = feat_name

        if scaler is not None:
            NotImplementedError()

        comparison = "<=" if not self.is_categorical else "="
        not_comparison = ">" if not self.is_categorical else "!="
        rounded_value = str(rule["threshold"]) if float_precision is None else round(rule["threshold"], float_precision)
        if scaler is not None:
            NotImplementedError()
        rule["textual_rule"] = f"{feat_name} {comparison} {rounded_value}\t{rule['samples']}"
        rule["blob_rule"] = f"{feat_name} {comparison} {rounded_value}"
        rule["graphviz_rule"] = {
            "label": f"{feat_name} {'\u2264' if not self.is_categorical else '='} {rounded_value}",
        }

        rule["not_textual_rule"] = f"{feat_name} {not_comparison} {rounded_value}"
        rule["not_blob_rule"] = f"{feat_name} {not_comparison} {rounded_value}"
        rule["not_graphviz_rule"] = {
            "label": f"{feat_name} {'>' if not self.is_categorical else '\u2260'} {rounded_value}"
        }

        return rule

    def node_to_dict(self):
        """
        Convert the classifier node to a dictionary representation.

        This method serializes the classifier's properties and configuration into a
        dictionary format, which can be used for persistence or knowledge transfer.

        Returns
        -------
        dict
            A dictionary containing all essential information about the classifier node,
            including the rule, classifier type, node statistics, and configuration parameters.
        """
        rule = self.get_rule(float_precision=None)

        rule["stump_type"] = self.__class__.__module__
        rule["samples"] = self.n_node_samples[0]
        rule["impurity"] = self.tree_.impurity[0]
     
      
        rule["args"] = {
            "is_oblique": self.is_oblique,
            "is_pivotal": self.is_pivotal,
            "unique_val_enum": self.unique_val_enum,
            "coefficients": self.coefficients,
            "num_pre_transformed" : self.num_pre_transformed,
            "cat_pre_transformed" : self.cat_pre_transformed,
            
            "distance_measure" : self.distance_measure #adding this for PT
            
        } | self.kwargs

        rule["split"] = {
            "args": {}
        }

        return rule
    
    @classmethod
    def dict_to_node(cls, node_dict, X = None):
        """
        Create a classifier node from a dictionary representation.

        This class method deserializes a dictionary back into a PivotTreeStumpClassifier
        instance, restoring its state and configuration.

        Parameters
        ----------
        node_dict : dict
            Dictionary containing the serialized classifier information.
        X : numpy.ndarray, optional
            Reference dataset that contains the pivot instance. Default is None.

        Returns
        -------
        PivotTreeStumpClassifier
            A reconstructed classifier instance.
        """
        self = cls()
        self.feature_original = np.zeros(3, dtype=int)
        self.threshold_original = np.zeros(3)
        self.n_node_samples = np.zeros(3, dtype=int)

        self.feature_original[0] = node_dict["feature_idx"]
        self.threshold_original[0] = node_dict["threshold"]
        self.n_node_samples[0] = node_dict["samples"]
        self.is_categorical = node_dict["is_categorical"]

        args = copy.deepcopy(node_dict["args"])
        self.is_oblique = args.pop("is_oblique")
        self.is_pivotal = args.pop("is_pivotal")
        self.unique_val_enum = args.pop("unique_val_enum")
        self.coefficients = args.pop("coefficients")
        self.kwargs = args
        
        self.distance_measure = args.pop("distance_measure")
        self.num_pre_transformed = args.pop("num_pre_transformed")
        self.cat_pre_transformed = args.pop("cat_pre_transformed")
        
        #X acts as a reference dataset for the instance id
        if X is not None:
            self.X_split_instance = X[int(node_dict["feature_idx"])]

        return self

    def export_graphviz(self, graph=None, columns_names=None, scaler=None, float_precision=3):
        """
        Export the classifier as a Graphviz digraph representation.

        This method is intended to visualize the classifier as a graph but is
        currently not implemented for PivotTreeStumpClassifier.

        Parameters
        ----------
        graph : pydot.Dot, optional
            An existing graph to add the classifier node to. Default is None.
        columns_names : list, optional
            List of column names for the features. Default is None.
        scaler : object, optional
            Scaling object used to transform feature values. Default is None.
        float_precision : int, optional
            Number of decimal places for floating point values. Default is 3.

        Raises
        ------
        NotImplementedError
            This method is not implemented for PivotTreeStumpClassifier.
        """
        raise NotImplementedError()



"""
Matrix to vector distance calculations for shapelet transforms.

This module provides optimized functions for computing various distance metrics
between a matrix of vectors and a single vector. These functions are critical for
efficient shapelet distance calculations in the RuleTree framework.
"""

import numpy as np
from numba import jit
from scipy.spatial import distance


@jit
def euclidean(matrix, vector):
    """
    Calculate the Euclidean distance between each row in a matrix and a vector.
    
    Parameters
    ----------
    matrix : numpy.ndarray
        A 2D numpy array where each row is a vector
    vector : numpy.ndarray
        A 1D numpy array to compute distances against
        
    Returns
    -------
    numpy.ndarray
        A 1D array containing the Euclidean distance between each row of the matrix and the vector
    """
    return np.sqrt(np.sum(np.power(matrix - vector, 2), axis=1))


@jit
def sqeuclidean(matrix, vector):
    """
    Calculate the squared Euclidean distance between each row in a matrix and a vector.
    
    Parameters
    ----------
    matrix : numpy.ndarray
        A 2D numpy array where each row is a vector
    vector : numpy.ndarray
        A 1D numpy array to compute distances against
        
    Returns
    -------
    numpy.ndarray
        A 1D array containing the squared Euclidean distance between each row of the matrix and the vector
    """
    return np.sum(np.power(matrix - vector, 2), axis=1)


@jit
def cityblock(matrix, vector):
    """
    Calculate the Manhattan (cityblock) distance between each row in a matrix and a vector.
    
    Parameters
    ----------
    matrix : numpy.ndarray
        A 2D numpy array where each row is a vector
    vector : numpy.ndarray
        A 1D numpy array to compute distances against
        
    Returns
    -------
    numpy.ndarray
        A 1D array containing the Manhattan distance between each row of the matrix and the vector
    """
    return np.sum(np.abs(matrix - vector), axis=1)


@jit(nopython=True)
def cosine(matrix, vector):
    """
    Calculate the cosine distance between each row in a matrix and a vector.
    
    Note: This function handles potential numerical issues by returning 0 for any
    NaN values that might occur during calculation.
    
    Parameters
    ----------
    matrix : numpy.ndarray
        A 2D numpy array where each row is a vector
    vector : numpy.ndarray
        A 1D numpy array to compute distances against
        
    Returns
    -------
    numpy.ndarray
        A 1D array containing the cosine distance between each row of the matrix and the vector
    """
    res = np.zeros(matrix.shape[0])

    for i in matrix.shape[0]:
        try:
            res[i] = 1 - np.dot(matrix[i], vector) / np.linalg.norm(matrix[i]) * np.linalg.norm(vector)
        except:
            res[i] = np.nan
            #warnings.warn("Er")

    return np.nan_to_num(res)


if __name__ == "__main__":
    matrix = np.array([
        [0, 1, 0],
        [0, 0, 0],
    ], dtype=np.float64)

    vector = np.array([0, -1, 0], dtype=np.float64)

    print("Euclidean:", euclidean(matrix, vector), [distance.euclidean(el, vector) for el in matrix])
    print("SqEuclidean:", sqeuclidean(matrix, vector), [distance.sqeuclidean(el, vector) for el in matrix])
    print("Cosine:", cosine(matrix, vector), [distance.cosine(el, vector) for el in matrix])
    print("CityBlock:", cityblock(matrix, vector), [distance.cityblock(el, vector) for el in matrix])



import numpy as np
import pandas as pd
from category_encoders import OrdinalEncoder

from scipy import linalg
from scipy.special import logsumexp
from sklearn.compose import ColumnTransformer, make_column_selector


def estimate_gaussian_covariances(X, nk, means, resp, reg_covar=1e-06):
    """Estimate the full covariance matrices.
    Parameters
    ----------
    resp : array-like of shape (n_samples, n_components)
    X : array-like of shape (n_samples, n_features)
    nk : array-like of shape (n_components,)
    means : array-like of shape (n_components, n_features)
    reg_covar : float
    Returns
    -------
    covariances : array, shape (n_components, n_features, n_features)
        The covariance matrix of the current components.
    """
    n_components, n_features = means.shape
    covariances = np.empty((n_components, n_features, n_features))
    for k in range(n_components):
        diff = X - means[k]
        covariances[k] = np.dot(resp[:, k] * diff.T, diff) / nk[k]
        covariances[k].flat[:: n_features + 1] += reg_covar
    return covariances


def estimate_parameters(X, resp, reg_covar=1e-06):
    nk = resp.sum(axis=0) + 10 * np.finfo(resp.dtype).eps
    means = np.dot(resp.T, X) / nk[:, np.newaxis]
    covariances = estimate_gaussian_covariances(X, nk, means, resp)
    return nk, means, covariances


def compute_precision_cholesky(covariances):
    """Compute the Cholesky decomposition of the precisions.
    Parameters
    ----------
    covariances : array-like
        The covariance matrix of the current components.
        The shape depends of the covariance_type.
    Returns
    -------
    precisions_cholesky : array-like
        The cholesky decomposition of sample precisions of the current
        components. The shape depends of the covariance_type.
    """
    n_components, n_features, _ = covariances.shape
    precisions_chol = np.empty((n_components, n_features, n_features))
    for k, covariance in enumerate(covariances):
        try:
            cov_chol = linalg.cholesky(covariance, lower=True)
        except linalg.LinAlgError:
            raise ValueError("Value error")
        precisions_chol[k] = linalg.solve_triangular(
            cov_chol, np.eye(n_features), lower=True
        ).T
    return precisions_chol


def estimate_weighted_log_prob(X, means, precisions_chol, weights):
    """Estimate the weighted log-probabilities, log P(X | Z) + log weights.
    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
    Returns
    -------
    weighted_log_prob : array, shape (n_samples, n_component)
    """
    return estimate_log_prob(X, means, precisions_chol) + estimate_log_weights(weights)


def estimate_log_prob(X, means, precisions_chol):
    """Estimate the log Gaussian probability.
    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
    means : array-like of shape (n_components, n_features)
    precisions_chol : array-like
    Returns
    -------
    log_prob : array, shape (n_samples, n_components)
    """
    n_samples, n_features = X.shape
    n_components, _ = means.shape
    # The determinant of the precision matrix from the Cholesky decomposition
    # corresponds to the negative half of the determinant of the full precision
    # matrix.
    # In short: det(precision_chol) = - det(precision) / 2
    log_det = compute_log_det_cholesky(precisions_chol, n_features)

    log_prob = np.empty((n_samples, n_components))
    for k, (mu, prec_chol) in enumerate(zip(means, precisions_chol)):
        y = np.dot(X, prec_chol) - np.dot(mu, prec_chol)
        log_prob[:, k] = np.sum(np.square(y), axis=1)

    # Since we are using the precision of the Cholesky decomposition,
    # `- 0.5 * log_det_precision` becomes `+ log_det_precision_chol`
    return -0.5 * (n_features * np.log(2 * np.pi) + log_prob) + log_det


def compute_log_det_cholesky(matrix_chol, n_features):
    """Compute the log-det of the cholesky decomposition of matrices.
    Parameters
    ----------
    matrix_chol : array-like
        Cholesky decompositions of the matrices.
    n_features : int
        Number of features.
    Returns
    -------
    log_det_precision_chol : array-like of shape (n_components,)
        The determinant of the precision matrix for each component.
    """
    n_components, _, _ = matrix_chol.shape
    log_det_chol = np.sum(
        np.log(matrix_chol.reshape(n_components, -1)[:, :: n_features + 1]), 1
    )

    return log_det_chol


def estimate_log_weights(weights):
    return np.log(weights)


def score_samples(X, means, precisions_chol, weights):
    """Compute the log-likelihood of each sample.
    Parameters
    ----------
    X : array-like of shape (n_samples, n_features)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.
    Returns
    -------
    log_prob : array, shape (n_samples,)
        Log-likelihood of each sample in `X` under the current model.
    """

    return logsumexp(
        estimate_weighted_log_prob(X, means, precisions_chol, weights), axis=1
    )


def score(X, n_components, labels):
    """Compute the per-sample average log-likelihood of the given data X.
    Parameters
    ----------
    X : array-like of shape (n_samples, n_dimensions)
        List of n_features-dimensional data points. Each row
        corresponds to a single data point.
    Returns
    -------
    log_likelihood : float
        Log-likelihood of `X` under the Gaussian mixture model.
    """

    n_samples = X.shape[0]

    resp = np.zeros((n_samples, n_components))
    resp[np.arange(n_samples), labels] = 1

    weights, means, covariances = estimate_parameters(X, resp, reg_covar=1e-06)
    weights /= n_samples

    precisions_chol = compute_precision_cholesky(covariances)

    return score_samples(X, means, precisions_chol, weights).mean()


def get_n_parameters(X, n_components):
    """Return the number of free parameters in the model."""
    n_features = X.shape[1]
    cov_params = n_components * n_features * (n_features + 1) / 2.0
    mean_params = n_features * n_components
    n_parameters = int(cov_params + mean_params + n_components - 1)
    return n_parameters


def bic(X, labels):
    """Bayesian information criterion for the current model on the input X.
    You can refer to this :ref:`mathematical section <aic_bic>` for more
    details regarding the formulation of the BIC used.
    Parameters
    ----------
    X : array of shape (n_samples, n_dimensions)
        The input samples.
    Returns
    -------
    bic : float
        The lower the better.
    """

    if X.dtype == np.dtype('O'):
        X = pd.DataFrame(X).infer_objects()
        ct = ColumnTransformer([("cat", OrdinalEncoder(), make_column_selector(dtype_include="object"))],
                               remainder='passthrough', verbose_feature_names_out=False, sparse_threshold=0, n_jobs=1)
        X = ct.fit_transform(X)

    n_records = X.shape[0]
    n_components = len(np.unique(labels))
    bic_score = score(X, n_components, labels)
    n_free_params = get_n_parameters(X, n_components)
    bic_val = -2 * bic_score * n_records + n_free_params * np.log(n_records)
    return bic_val



"""Factor Analysis of Mixed Data (FAMD)"""
import numpy as np
import pandas as pd

from . import mfa


class FAMD(mfa.MFA):

    def __init__(self, n_components=2, n_iter=2, copy=False, check_input=True, random_state=None,
                 engine='auto'):
        super().__init__(
            groups=None,
            normalize=True,
            n_components=n_components,
            n_iter=n_iter,
            copy=copy,
            check_input=check_input,
            random_state=random_state,
            engine=engine
        )

    def fit(self, X, y=None):

        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        # Separate numerical columns from categorical columns
        num_cols = X.select_dtypes(np.number).columns
        cat_cols = X .select_dtypes(include=['object','category']).columns

        # Make one group per variable type
        self.groups = {}
        if len(num_cols):
            self.groups['Numerical'] = num_cols
        else:
            raise ValueError("FAMD works with categorical and numerical data but " +
                             "you only have categorical data; you should consider using MCA")
        if len(cat_cols):
            self.groups['Categorical'] = cat_cols
        else:
            raise ValueError("FAMD works with categorical and numerical data but " +
                             "you only have numerical data; you should consider using PCA")

        return super().fit(X)
    
    def fit_transform(self,X,y=None):
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X).infer_objects()
            X.columns = [str(x) for x in X.columns]

        # Separate numerical columns from categorical columns
        num_cols = X.select_dtypes(np.number).columns
        cat_cols = X .select_dtypes(include=['object','category']).columns

        # Make one group per variable type
        self.groups = {}
        if len(num_cols):
            self.groups['Numerical'] = num_cols
        else:
            raise ValueError("FAMD works with categorical and numerical data but " +
                             "you only have categorical data; you should consider using MCA")
        if len(cat_cols):
            self.groups['Categorical'] = cat_cols
        else:
            raise ValueError("FAMD works with categorical and numerical data but " +
                             "you only have numerical data; you should consider using PCA")
        return super().fit_transform(X)
        

  










"""Multiple Correspondence Analysis (MCA)"""

import numpy as np
from sklearn import utils

from . import ca
from . import one_hot


class MCA(ca.CA):

    def fit(self, X, y=None):
        if self.check_input:
            utils.check_array(X, dtype=[str, np.number])
            
        n_initial_columns = X.shape[1]

        # One-hot encode the data
        self.one_hot_ = one_hot.OneHotEncoder().fit(X)
        
        _X_t=  self.one_hot_.transform(X) 
        
        _0_freq_serie= (_X_t == 0).sum(axis=0)/ len(_X_t)
        
        self._usecols=_0_freq_serie[_0_freq_serie < 0.99].index
        #print('MCA PROCESS ELIMINATED {0}  COLUMNS SINCE THEIR MISS_RATES >= 99%'.format( _X_t.shape[1] - len(self._usecols) ))
        
        n_new_columns = len(self._usecols)
        self.total_inertia_ = (n_new_columns - n_initial_columns) / n_initial_columns
        # Apply CA to the indicator matrix
        super().fit(_X_t.loc[:,self._usecols])

        return self

    def _transform(self, X):
        return super()._transform(self.one_hot_.transform(X).loc[:,self._usecols])



    def transform(self, X):
        """Computes the row principal coordinates of a dataset."""
        utils.validation.check_is_fitted(self, 'singular_values_')
        if self.check_input:
            utils.check_array(X, dtype=[str, np.number])
        return self._transform(X)




"""
This file defines an abstract base class for splitting strategies in RuleTree models.
It provides a framework that concrete implementations of split strategies must adhere to.
By inheriting from Python's Abstract Base Class (ABC), it enforces the implementation of necessary methods,
ensuring consistency and compatibility across different splitting algorithms used within RuleTree models.
"""
from abc import ABC, abstractmethod

class RuleTreeBaseSplit(ABC):
    """
    Abstract base class for split strategies in RuleTree models.

    This class defines the interface that all concrete implementations of split strategies must follow.
    It is designed to be used within RuleTree models to ensure consistent and compatible splitting behavior
    across different algorithms. By inheriting from Python's Abstract Base Class (ABC), it enforces the
    implementation of necessary methods, such as the constructor that initializes with a machine learning task type.

    Attributes:
        ml_task (str): The type of machine learning task for which the split strategy is being used
                       (e.g., 'classification', 'regression').
    """

    @abstractmethod
    def __init__(self, ml_task):
        """
        Initializes the RuleTreeBaseSplit with a specific machine learning task type.

        Args:
            ml_task (str): The type of machine learning task for which the split strategy is being used.
        """
        self.ml_task = ml_task



import numpy as np
import pandas as pd
from scipy.stats import pearsonr


def _pearsonr(A,B,threshold=0.05):
    co_,p_value=pearsonr(A,B)
    return co_ if p_value <=threshold else  np.nan
    
    
    
def  df_correlation(DF_C,DF_INDEX):
    if not isinstance(DF_C,(pd.DataFrame,pd.SparseDataFrame)):
        DF_C=pd.DataFrame(DF_C)
    if not isinstance(DF_INDEX,(pd.DataFrame,pd.SparseDataFrame)):
        DF_INDEX=pd.DataFrame(DF_INDEX)		
    _data={}
    for col in DF_C:
        if isinstance(DF_C[col],pd.SparseSeries) :
            _c= DF_C[col].values.to_dense()
        else:
            _c= DF_C[col].values
        _embed_dict={}
        for index in  DF_INDEX:
            if isinstance(DF_INDEX[index],pd.SparseSeries) :
                _i= DF_INDEX[index].values.to_dense()
            else:
                _i= DF_INDEX[index].values
            _embed_dict[index]=_pearsonr(_c,_i)
        _data[col]=_embed_dict
            
    return pd.DataFrame(_data )    



from random import random
import numpy as np
import sklearn.base
from sklearn.ensemble import BaggingClassifier
from RuleTree.utils.data_utils import _iterative_mean

from RuleTree import RuleTreeClassifier
from RuleTree.base.RuleTreeBase import RuleTreeBase
from sklearn.base import ClassifierMixin

from RuleTree.stumps.classification.DecisionTreeStumpClassifier import DecisionTreeStumpClassifier

class RuleForestClassifier(BaggingClassifier, RuleTreeBase):
    """
    Random forest implementation based on Rule Trees.

    RuleForestClassifier combines multiple RuleTreeClassifiers to improve prediction accuracy and
    control overfitting. Each tree in the forest uses a random subset of features and samples for training.

    This implementation extends scikit-learn's BaggingClassifier and adds functionalities specific to
    rule-based classifiers.
    """ 

    def __init__(self,
                 n_estimators=100,
                 criterion='gini',
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.0,
                 min_impurity_decrease=0.0,
                 max_leaf_nodes=float("inf"),
                 class_weight=None,
                 ccp_alpha=0.0,
                 prune_useless_leaves=False,
                 splitter='best',
                 *,
                 max_samples=None,
                 max_features=1.0,
                 bootstrap=True,
                 bootstrap_features = True,
                 oob_score=False,
                 warm_start=False,
                 custom_estimator:sklearn.base.ClassifierMixin=None,
                 n_jobs=None,
                 random_state=None,
                 base_stumps = None,
                 distance_matrix = None,
                 distance_measure = None,
                 stump_selection = 'best',
                 verbose=0):
        """
        Initialize a RuleForestClassifier.

        Parameters
        ----------
        n_estimators : int, default=100
            The number of trees in the forest.

        criterion : {'gini', 'entropy'}, default='gini'
            Function to measure the quality of a split.

        max_depth : int or None, default=None
            The maximum depth of the tree. If None, then nodes are expanded until all leaves
            are pure or until all leaves contain less than min_samples_split samples.

        min_samples_split : int or float, default=2
            The minimum number of samples required to split an internal node.

        min_samples_leaf : int or float, default=1
            The minimum number of samples required to be at a leaf node.

        min_weight_fraction_leaf : float, default=0.0
            The minimum weighted fraction of the sum total of weights required to be at a leaf node.

        min_impurity_decrease : float, default=0.0
            A node will be split if this split induces a decrease of the impurity greater than or equal
            to this value.

        max_leaf_nodes : int or None, default=float("inf")
            Grow trees with max_leaf_nodes in best-first fashion.

        class_weight : dict, list of dict or "balanced", default=None
            Weights associated with classes.

        ccp_alpha : non-negative float, default=0.0
            Complexity parameter used for Minimal Cost-Complexity Pruning.

        prune_useless_leaves : bool, default=False
            Whether to prune useless leaves that do not contribute to predictions.

        splitter : {"best", "random", "hybrid_forest"}, default="best"
            The strategy used to choose the split at each node.

        max_samples : int or float, default=None
            The number of samples to draw from X to train each base estimator.

        max_features : int, float, str or None, default=1.0
            The number of features to consider when looking for the best split.

        bootstrap : bool, default=True
            Whether samples are drawn with replacement.

        bootstrap_features : bool, default=True
            Whether features are drawn with replacement.

        oob_score : bool, default=False
            Whether to use out-of-bag samples to estimate the generalization error.

        warm_start : bool, default=False
            When set to True, reuse the solution of the previous call to fit.

        custom_estimator : sklearn.base.ClassifierMixin, default=None
            The base estimator to use. If None, RuleTreeClassifier is used.

        n_jobs : int or None, default=None
            The number of jobs to run in parallel.

        random_state : int, RandomState instance or None, default=None
            Controls the random resampling of the original dataset.

        base_stumps : list, default=None
            List of base stumps to use for building the trees.

        distance_matrix : array-like, default=None
            Pre-computed distance matrix to use for certain types of stumps.

        distance_measure : callable, default=None
            Function to compute distances between samples.

        stump_selection : {'best', 'random'}, default='best'
            Strategy to select stumps when building trees.

        verbose : int, default=0
            Controls the verbosity when fitting and predicting.
        """

        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.min_impurity_decrease = min_impurity_decrease
        self.max_leaf_nodes = max_leaf_nodes
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.prune_useless_leaves = prune_useless_leaves
        self.splitter = splitter
        self.max_samples = max_samples
        self.max_features = max_features
        self.bootstrap = bootstrap
        self.bootstrap_features = bootstrap_features
        self.oob_score = oob_score
        self.warm_start = warm_start
        self.custom_estimator = custom_estimator
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.base_stumps = base_stumps
        self.distance_matrix = distance_matrix
        self.distance_measure = distance_measure
        self.stump_selection= stump_selection
    
               

    def fit(self, X: np.ndarray, y:np.ndarray, sample_weight=None, **kwargs):
        """
        Build a forest of rule trees from the training set (X, y).

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels).

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted.

        **kwargs : dict
            Additional parameters passed to the base estimator's fit method.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        if self.max_features is None:
            self.max_features = X.shape[1]
            

        if type(self.max_features) is str:
            if self.max_features == "sqrt":
                self.max_features = int(np.sqrt(X.shape[1]))
            elif self.max_features == "log2":
                self.max_features = int(np.log2(X.shape[1]))

        base_estimator = RuleTreeClassifier if self.custom_estimator is None else self.custom_estimator
        splitter = .5 if self.splitter == 'hybrid_forest' else self.splitter
        
        if type(splitter) is float:
            base_estimator = RuleTreeClassifier_choosing_splitter_randomly
            
        if self.base_stumps is None:
            self.base_stumps = [DecisionTreeStumpClassifier(
                                max_depth=1,
                                criterion=self.criterion,
                                splitter=self.splitter,
                                min_samples_split=self.min_samples_split,
                                min_samples_leaf = self.min_samples_leaf,
                                min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                                max_features=self.max_features,
                                random_state=self.random_state,
                                min_impurity_decrease=self.min_impurity_decrease,
                                class_weight=self.class_weight,
                                ccp_alpha=self.ccp_alpha,
                              #  monotonic_cst = self.monotonic_cst
                                )]
        
        else:
            for stump in self.base_stumps:
                if stump.__class__.__module__.split('.')[-1] in [
                    'PivotTreeStumpClassifier',
                    'MultiplePivotTreeStumpClassifier',
                    'ObliquePivotTreeStumpClassifier',
                    'MultipleObliquePivotTreeStumpClassifier'
                ]:
                    
                    base_estimator = ForestEstimatorPivotClassifier
                    break
                
            
     
        super().__init__(estimator=base_estimator(criterion=self.criterion,
                                                  max_depth=self.max_depth,
                                                  min_samples_split=self.min_samples_split,
                                                  min_samples_leaf=self.min_samples_leaf,
                                                  min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                                                  min_impurity_decrease=self.min_impurity_decrease,
                                                  random_state=self.random_state,
                                                  max_leaf_nodes=self.max_leaf_nodes,
                                                  class_weight=self.class_weight,
                                                  ccp_alpha=self.ccp_alpha,
                                                  prune_useless_leaves=self.prune_useless_leaves,
                                                  splitter=self.splitter,
                                                  base_stumps = self.base_stumps,
                                                  distance_measure = self.distance_measure,
                                                  distance_matrix = self.distance_matrix,
                                                  stump_selection= self.stump_selection
                        
                                                  ),
                         n_estimators=self.n_estimators,
                         max_samples=X.shape[0] if self.max_samples is None else self.max_samples,
                         max_features=self.max_features,
                         bootstrap=self.bootstrap,
                         bootstrap_features=self.bootstrap_features,
                         oob_score=self.oob_score,
                         warm_start=self.warm_start,
                         n_jobs=self.n_jobs,
                         random_state=self.random_state,
                         verbose=self.verbose)

        return super().fit(X, y, sample_weight=sample_weight, **kwargs)
     
    def local_interpretation(self, X, joint_contribution = False):
        """
        Provide local interpretation of the model for the given samples.

        This method calculates feature contributions to the prediction for each sample in X.
        It aggregates interpretations from all trees in the forest.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to interpret.

        joint_contribution : bool, default=False
            If True, returns joint contributions of features.
            If False, returns individual feature contributions.

        Returns
        -------
        tuple
            A tuple containing:
            - predictions: array of shape (n_samples,) - the mean prediction
            - biases: array of shape (n_samples,) - the bias term (prior probability)
            - contributions: array or list of dicts - feature contributions to the prediction
        """

        if joint_contribution:
            biases = []
            contributions = []
            predictions = []
            
            for tree in self.estimators_:
                pred, bias, contribution = tree.local_interpretation(X, joint_contribution=joint_contribution)
                biases.append(bias)
                contributions.append(contribution)
                predictions.append(pred)
                
            total_contributions = []
            
            for i in range(len(X)):
                contr = {}
                for j, dct in enumerate(contributions):
                    for k in set(dct[i]).union(set(contr.keys())):
                        contr[k] = (contr.get(k, 0)*j + dct[i].get(k,0) ) / (j+1)

                total_contributions.append(contr)    
                
            for i, item in enumerate(contribution):
                total_contributions[i]
                sm = sum([v for v in contribution[i].values()])
                    

            
            return (np.mean(predictions, axis=0), np.mean(biases, axis=0),
                total_contributions)
        else:
            mean_pred = None
            mean_bias = None
            mean_contribution = None
            
            for i, tree in enumerate(self.estimators_):
                pred, bias, contribution = tree.local_interpretation(X)
                
                if i < 1: # first iteration
                    mean_bias = bias
                    mean_contribution = contribution
                    mean_pred = pred
                else:
                    mean_bias = _iterative_mean(i, mean_bias, bias)
                    mean_contribution = _iterative_mean(i, mean_contribution, contribution)
                    mean_pred = _iterative_mean(i, mean_pred, pred)

            return mean_pred, mean_bias, mean_contribution

        

class RuleTreeClassifier_choosing_splitter_randomly(RuleTreeClassifier):
    """
    A variant of RuleTreeClassifier that randomly chooses between 'best' and 'random' splitters.

    This class is used when the splitter parameter is set as a float between 0 and 1.
    The float value determines the probability of using the 'random' splitter.

    Parameters
    ----------
    splitter : float
        Probability value between 0 and 1 for choosing the splitter strategy.

    **kwargs : dict
        Additional parameters passed to RuleTreeClassifier.
    """
    def __init__(self, splitter, **kwargs):
        if random() < splitter:
            if random() < splitter:
                splitter = 'random'
            else:
                splitter = 'best'
        kwargs["splitter"] = splitter
        super().__init__(**kwargs)
        
class ForestEstimatorPivotClassifier(RuleTreeClassifier):
    """
    A specialized RuleTreeClassifier for use with pivot-based stumps in forest ensembles.

    This classifier is automatically selected when certain types of pivot tree stumps
    are used as base stumps in the RuleForestClassifier.

    Parameters
    ----------
    max_leaf_nodes : int, default=float('inf')
        Grow trees with max_leaf_nodes in best-first fashion.

    min_samples_split : int, default=2
        The minimum number of samples required to split an internal node.

    max_depth : int, default=float('inf')
        The maximum depth of the tree.

    prune_useless_leaves : bool, default=False
        Whether to prune useless leaves that do not contribute to predictions.

    base_stumps : estimator or list of estimators, default=None
        The base stumps to use when building the tree.

    stump_selection : {'best', 'random'}, default='random'
        The strategy to use when selecting stumps.

    random_state : int, RandomState instance or None, default=None
        Controls the randomness of the estimator.

    criterion : str, default='gini'
        The function to measure the quality of a split.

    splitter : str, default='best'
        The strategy used to choose the split at each node.

    min_samples_leaf : int, default=1
        The minimum number of samples required to be at a leaf node.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights required to be at a leaf node.

    max_features : int, float, str or None, default=None
        The number of features to consider when looking for the best split.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity greater than this value.

    class_weight : dict, list of dict or "balanced", default=None
        Weights associated with classes.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning.

    monotonic_cst : array-like of int of shape (n_features), default=None
        Monotonicity constraints for each feature.

    distance_matrix : array-like, default=None
        Pre-computed distance matrix.

    distance_measure : callable, default=None
        Function to compute distances between samples.
    """
    def __init__(self,
                 max_leaf_nodes=float('inf'),
                 min_samples_split=2,
                 max_depth=float('inf'),
                 prune_useless_leaves=False,
                 base_stumps: ClassifierMixin | list = None,
                 stump_selection: str = 'random',
                 random_state=None,

                 criterion='gini',
                 splitter='best',
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.0,
                 max_features=None,
                 min_impurity_decrease=0.0,
                 class_weight=None,
                 ccp_alpha=0.0,
                 monotonic_cst=None,
                 distance_matrix = None,
                 distance_measure = None
                 
                 ):
        
        super().__init__(max_leaf_nodes=max_leaf_nodes,
                         min_samples_split=min_samples_split,
                         max_depth=max_depth,
                         prune_useless_leaves=prune_useless_leaves,
                         base_stumps=base_stumps,
                         stump_selection=stump_selection,
                         random_state=random_state)

        self.max_depth = max_depth
        self.criterion = criterion
        self.splitter = splitter
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.random_state = random_state
        self.min_impurity_decrease = min_impurity_decrease
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.monotonic_cst = monotonic_cst
        self.distance_matrix = distance_matrix    
        self.distance_measure = distance_measure

    
    def fit(self, X: np.array, y: np.array=None, **kwargs):
        """
        Build a forest estimator pivot classifier from the training set (X, y).

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.

        y : array-like of shape (n_samples,)
            The target values (class labels).

        **kwargs : dict
            Additional parameters passed to the base fit method.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        super().fit(X, y, **kwargs)



from sklearn.ensemble import AdaBoostClassifier

from RuleTree import RuleTreeClassifier
from RuleTree.base.RuleTreeBase import RuleTreeBase


class RuleTreeAdaBoostClassifier(AdaBoostClassifier, RuleTreeBase):
    """An AdaBoost classifier that uses RuleTreeClassifier as the base estimator.
    
    This implementation combines the AdaBoost algorithm with RuleTree-based weak learners.
    It inherits from both AdaBoostClassifier and RuleTreeBase, providing boosting
    capabilities with rule-based decision trees.
    
    AdaBoost builds a strong classifier by combining multiple weak classifiers (RuleTrees)
    sequentially, where each subsequent model tries to correct errors made by 
    the previous ones by adjusting the weights of incorrectly classified examples.
    
    Parameters
    ----------
    n_estimators : int, default=50
        The maximum number of estimators at which boosting is terminated.
        In case of perfect fit, the learning procedure is stopped early.
    
    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node.
        
    prune_useless_leaves : bool, default=False
        Whether to prune leaves that do not improve the performance.
        
    random_state : int, RandomState instance or None, default=None
        Controls the random seed given to each RuleTree estimator and the bootstrap
        sample selection for building the ensemble.
        
    criterion : {"gini", "entropy"}, default="gini"
        The function to measure the quality of a split.
        
    splitter : {"best", "random"}, default="best"
        The strategy used to choose the split at each node.
        
    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        
    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights required to be at a leaf node.
        
    max_features : int, float or {"auto", "sqrt", "log2"}, default=None
        The number of features to consider when looking for the best split.
        
    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity greater than or equal
        to this value.
        
    class_weight : dict, list of dict or "balanced", default=None
        Weights associated with classes.
        
    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning.
        
    monotonic_cst : array-like of int of shape (n_features), default=None
        Monotonicity constraints for each feature.
        
    learning_rate : float, default=1.0
        Learning rate shrinks the contribution of each classifier by `learning_rate`.
        
    algorithm : {'SAMME', 'SAMME.R'}, default='SAMME'
        The boosting algorithm to use.
    
    Attributes
    ----------
    estimator_ : RuleTreeClassifier
        The base estimator from which the ensemble is grown.
        
    estimators_ : list of RuleTreeClassifier
        The collection of fitted RuleTree sub-estimators.
        
    classes_ : ndarray of shape (n_classes,)
        The classes labels.
        
    n_classes_ : int
        The number of classes.
        
    See Also
    --------
    RuleTreeClassifier : Base rule tree classifier.
    sklearn.ensemble.AdaBoostClassifier : Original scikit-learn AdaBoost implementation.
    """
    
    def __init__(self,
                 n_estimators=50,
                 min_samples_split=2,
                 prune_useless_leaves=False,
                 random_state=None,
                 criterion='gini',
                 splitter='best',
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.0,
                 max_features=None,
                 min_impurity_decrease=0.0,
                 class_weight=None,
                 ccp_alpha=0.0,
                 monotonic_cst=None,
                 *,
                 learning_rate=1.0,
                 algorithm='SAMME'
                 ):
        """Initialize the RuleTreeAdaBoostClassifier.
        
        Parameters are the same as in the class docstring.
        """
        self.min_samples_split = min_samples_split
        self.prune_useless_leaves = prune_useless_leaves
        self.random_state = random_state
        self.criterion = criterion
        self.splitter = splitter
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.min_impurity_decrease = min_impurity_decrease
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.monotonic_cst = monotonic_cst
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.algorithm = algorithm

        estimator = RuleTreeClassifier(min_samples_split=min_samples_split,
                                         max_depth=3, #stump
                                         prune_useless_leaves=prune_useless_leaves,
                                         random_state=random_state,

                                         criterion=criterion,
                                         splitter=splitter,
                                         min_samples_leaf=min_samples_leaf,
                                         min_weight_fraction_leaf=min_weight_fraction_leaf,
                                         max_features=max_features,
                                         min_impurity_decrease=min_impurity_decrease,
                                         class_weight=class_weight,
                                         ccp_alpha=ccp_alpha,
                                         monotonic_cst=monotonic_cst
                                         )

        super().__init__(
            estimator=estimator,
            n_estimators=n_estimators, learning_rate=learning_rate, algorithm=algorithm, random_state=random_state
        )



import copy
import random
import warnings

import numpy as np
import psutil
import tempfile312
from matplotlib import pyplot as plt
from numba import UnsupportedError

from RuleTree.stumps.classification import ProximityTreeStumpClassifier
from RuleTree.stumps.regression import DecisionTreeStumpRegressor
from RuleTree.utils.define import DATA_TYPE_TS
from RuleTree.utils.shapelet_transform.Shapelets import Shapelets


class ProximityTreeStumpRegressor(DecisionTreeStumpRegressor):
    """A decision stump regressor for time series data using shapelet-based proximity measures.

    This regressor transforms time series data using shapelets (subsequences of time series)
    and creates regression rules based on the distances between the input time series and
    extracted shapelets. It predicts continuous target values rather than class labels.

    Parameters
    ----------
    n_shapelets : int, default=cpu_count*2
        The number of shapelets to be used for regression.
    n_shapelets_for_selection : int or str, default=500
        Number of shapelets to select or 'stratified' for stratified selection.
    n_ts_for_selection_per_class : int, default=100
        Number of time series per class to use for shapelet extraction.
    sliding_window : int, default=50
        The length of the sliding window used to extract shapelets from time series.
    selection : str, default='mi_clf'
        Method for shapelet selection. Options: 'random', 'mi_clf', 'mi_reg', 'cluster'.
    distance : str, default='euclidean'
        Distance metric used for comparing shapelets to time series.
    mi_n_neighbors : int, default=100
        Number of neighbors for mutual information calculation.
    random_state : int, default=42
        Seed for random number generation for reproducibility.
    n_jobs : int, default=1
        Number of parallel jobs to run for shapelet extraction.
    **kwargs
        Additional parameters for the base DecisionTreeStumpRegressor.
    """
    def __init__(self,
                 n_shapelets=psutil.cpu_count(logical=False) * 2,
                 n_shapelets_for_selection=500,  # int, inf, or 'stratified'
                 n_ts_for_selection_per_class=100,  # int, inf
                 sliding_window=50,
                 selection='mi_clf',  # random, mi_clf, mi_reg, cluster
                 distance='euclidean',
                 mi_n_neighbors=100,
                 random_state=42, n_jobs=1,
                 **kwargs):
        self.n_shapelets = n_shapelets
        self.n_shapelets_for_selection = n_shapelets_for_selection
        self.n_ts_for_selection_per_class = n_ts_for_selection_per_class
        self.sliding_window = sliding_window
        self.selection = selection
        self.distance = distance
        self.mi_n_neighbors = mi_n_neighbors
        self.random_state = random_state
        self.n_jobs = n_jobs

        if "max_depth" in kwargs and kwargs["max_depth"] > 1:
            warnings.warn("max_depth must be 1")

        kwargs["max_depth"] = 1

        if selection not in ['random', 'mi_clf', 'cluster']:
            raise ValueError("'selection' must be 'random', 'mi_clf' or 'cluster'")

        super().__init__(**kwargs)

        self.kwargs |= {
            "n_shapelets": n_shapelets,
            "n_shapelets_for_selection": n_shapelets_for_selection,
            "n_ts_for_selection_per_class": n_ts_for_selection_per_class,
            "sliding_window": sliding_window,
            "selection": selection,
            "distance": distance,
            "mi_n_neighbors": mi_n_neighbors,
            "random_state": random_state,
            "n_jobs": n_jobs,
        }

    def fit(self, X, y, idx=None, context=None, sample_weight=None, check_input=True):
        """Fit the ProximityTreeStumpRegressor on the training data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_timepoints)
            The time series data.
        y : array-like of shape (n_samples,)
            The target continuous values.
        idx : array-like or slice, default=None
            Indices of the samples to be used for training.
        context : object, default=None
            Additional context information (not used).
        sample_weight : array-like, default=None
            Sample weights (not supported for this regressor).
        check_input : bool, default=True
            Whether to check the input parameters.

        Returns
        -------
        self : object
            Fitted estimator.

        Raises
        ------
        UnsupportedError
            If sample_weight is provided.
        """
        if idx is None:
            idx = slice(None)
        X = X[idx]
        y = y[idx]

        self.y_lims = [X.min(), X.max()]

        random.seed(self.random_state)
        if sample_weight is not None:
            raise UnsupportedError(f"sample_weight is not supported for {self.__class__.__name__}")

        self.st = Shapelets(n_shapelets=self.n_shapelets,
                            n_shapelets_for_selection=self.n_shapelets_for_selection,
                            n_ts_for_selection_per_class=self.n_ts_for_selection_per_class,
                            sliding_window=self.sliding_window,
                            selection=self.selection,
                            distance=self.distance,
                            mi_n_neighbors=self.mi_n_neighbors,
                            random_state=random.randint(0, 2**32-1),
                            n_jobs=self.n_jobs
                            )

        X_dist = self.st.fit_transform(X, y)
        actual_n_shapelets = X_dist.shape[1]
        X_bool = np.zeros((X.shape[0], actual_n_shapelets*(actual_n_shapelets-1)), dtype=bool)

        c = 0

        for i in range(actual_n_shapelets):
            for j in range(i+1, actual_n_shapelets):
                X_bool[:, c] = X_dist[:, i] <= X_dist[:, j]
                c += 1

        return super().fit(X_bool, y=y, sample_weight=sample_weight, check_input=check_input)

    def apply(self, X, check_input=False):
        """Apply the decision tree stump to X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_timepoints)
            The time series data for regression.
        check_input : bool, default=False
            Whether to check the input parameters.

        Returns
        -------
        X_leaves : array-like of shape (n_samples,)
            For each datapoint x in X, returns the leaf node it ends up in.
        """
        self.y_lims = [min(self.y_lims[0], X.min()), min(self.y_lims[1], X.max())]
        X_dist = self.st.transform(X)
        actual_n_shapelets = X_dist.shape[1]
        X_bool = np.zeros((X.shape[0], actual_n_shapelets * (actual_n_shapelets - 1)), dtype=bool)

        c = 0

        for i in range(actual_n_shapelets):
            for j in range(i + 1, actual_n_shapelets):
                X_bool[:, c] = X_dist[:, i] <= X_dist[:, j]
                c += 1


        return super().apply(X_bool, check_input=check_input)

    def supports(self, data_type):
        """Check if the regressor supports the given data type.

        Parameters
        ----------
        data_type : int
            Data type identifier.

        Returns
        -------
        bool
            True if the data type is supported, False otherwise.
        """
        return data_type in [DATA_TYPE_TS]


    def get_rule(self, columns_names=None, scaler=None, float_precision: int | None = 3):
        """Generate a human-readable rule representation of the decision stump.

        Delegates to the ProximityTreeStumpClassifier implementation.

        Parameters
        ----------
        columns_names : list, default=None
            Names of the features (not used for time series).
        scaler : object, default=None
            Scaler used for normalization (not supported).
        float_precision : int or None, default=3
            Number of decimal places to use for floating-point values.

        Returns
        -------
        dict
            Dictionary containing textual and visual representations of the rule.
        """
        return ProximityTreeStumpClassifier.get_rule(self, columns_names, scaler, float_precision)

    def node_to_dict(self):
        """Convert the decision stump to a dictionary representation.

        Delegates to the ProximityTreeStumpClassifier implementation.

        Returns
        -------
        dict
            Dictionary representing the decision stump, including shapelet information.
        """
        return ProximityTreeStumpClassifier.node_to_dict(self)

    @classmethod
    def dict_to_node(cls, node_dict, X=None):
        """Convert a dictionary representation back to a decision stump.

        This method is used for model deserialization.

        Parameters
        ----------
        node_dict : dict
            Dictionary containing the serialized decision stump.
        X : array-like, default=None
            The time series data (not used).

        Returns
        -------
        ProximityTreeStumpRegressor
            The reconstructed decision stump regressor.
        """
        self = cls(
            n_shapelets=node_dict["n_shapelets"],
            n_shapelets_for_selection=node_dict["n_shapelets_for_selection"],
            n_ts_for_selection_per_class=node_dict["n_ts_for_selection_per_class"],
            sliding_window=node_dict["sliding_window"],
            selection=node_dict["selection"],
            distance=node_dict["distance"],
            mi_n_neighbors=node_dict["mi_n_neighbors"],
            random_state=node_dict["random_state"],
            n_jobs=node_dict["n_jobs"]
        )

        self.st = Shapelets(
            n_shapelets=node_dict["n_shapelets"],
            n_shapelets_for_selection=node_dict["n_shapelets_for_selection"],
            n_ts_for_selection_per_class=node_dict["n_ts_for_selection_per_class"],
            sliding_window=node_dict["sliding_window"],
            selection=node_dict["selection"],
            distance=node_dict["distance"],
            mi_n_neighbors=node_dict["mi_n_neighbors"],
            random_state=node_dict["random_state"],
            n_jobs=node_dict["n_jobs"]
        )

        self.st.shapelets = np.array(node_dict["shapelets"])

        self.feature_original = np.zeros(3, dtype=int)
        self.threshold_original = np.zeros(3)
        self.n_node_samples = np.zeros(3, dtype=int)

        self.y_lims = node_dict["y_lims"]

        self.feature_original[0] = node_dict["feature_idx"]
        self.threshold_original[0] = node_dict["threshold"]
        self.n_node_samples[0] = node_dict["samples"]
        self.is_categorical = node_dict["is_categorical"]

        args = copy.deepcopy(node_dict["args"])
        self.is_oblique = args.pop("is_oblique")
        self.is_pivotal = args.pop("is_pivotal")
        self.unique_val_enum = args.pop("unique_val_enum")
        self.coefficients = args.pop("coefficients")
        self.kwargs = args

        return self



"""
Module containing the RuleTreeClassifier class.

The RuleTreeClassifier class extends the RuleTree to implement classification-specific
functionality, including methods for handling class probabilities and medoid computation.
"""

import heapq

import numpy as np
import sklearn
from sklearn import tree
from sklearn.base import ClassifierMixin
import copy

from RuleTree.stumps.classification.MultiplePivotTreeStumpClassifier import MultiplePivotTreeStumpClassifier
from RuleTree.stumps.classification.PivotTreeStumpClassifier import PivotTreeStumpClassifier
from RuleTree.stumps.classification.ObliquePivotTreeStumpClassifier import ObliquePivotTreeStumpClassifier
from RuleTree.tree.RuleTree import RuleTree
from RuleTree.tree.RuleTreeNode import RuleTreeNode
from RuleTree.stumps.classification.DecisionTreeStumpClassifier import DecisionTreeStumpClassifier
from RuleTree.utils.data_utils import calculate_mode, get_info_gain

from RuleTree.utils.utils_decoding import configure_non_cat_split, configure_cat_split
from RuleTree.utils.utils_decoding import set_node_children, simplify_decode
from sklearn.metrics import pairwise_distances


class RuleTreeClassifier(RuleTree, ClassifierMixin):
    """
    RuleTree implementation for classification tasks.

    This class extends the RuleTree to include classification-specific methods
    and attributes, such as handling class probabilities and computing medoids.

    Attributes:
        criterion (str): Splitting criterion ('gini' or 'entropy').
        splitter (str): Split strategy ('best' or 'random').
        min_samples_leaf (int): Minimum samples required in a leaf node.
        distance_measure (str): Distance metric for medoid computation.
    """

    def __init__(self,
                 max_leaf_nodes=float('inf'),
                 min_samples_split=2,
                 max_depth=float('inf'),
                 prune_useless_leaves=False,
                 base_stumps: ClassifierMixin | list = None,
                 stump_selection: str = 'random',
                 random_state=None,

                 criterion='gini',
                 splitter='best',
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.0,
                 max_features=None,
                 min_impurity_decrease=0.0,
                 class_weight=None,
                 ccp_alpha=0.0,
                 monotonic_cst=None,
                 distance_matrix=None,
                 distance_measure=None

                 ):
        """
        Initialize the RuleTreeClassifier.

        Args:
            max_leaf_nodes (int): Maximum number of leaf nodes.
            min_samples_split (int): Minimum samples required to split a node.
            max_depth (int): Maximum depth of the tree.
            prune_useless_leaves (bool): Whether to prune useless leaves.
            base_stumps (list): List of base stumps or their configurations.
            stump_selection (str): Stump selection strategy ('random' or 'best').
            random_state (int): Random seed for reproducibility.
            criterion (str): Splitting criterion ('gini' or 'entropy').
            splitter (str): Split strategy ('best' or 'random').
            min_samples_leaf (int): Minimum samples required in a leaf node.
            distance_measure (str): Distance metric for medoid computation.
        """
        if base_stumps is None:
            base_stumps = DecisionTreeStumpClassifier(
                max_depth=1,
                criterion=criterion,
                splitter=splitter,
                min_samples_split=min_samples_split,
                min_samples_leaf=min_samples_leaf,
                min_weight_fraction_leaf=min_weight_fraction_leaf,
                max_features=max_features,
                random_state=random_state,
                min_impurity_decrease=min_impurity_decrease,
                class_weight=class_weight,
                ccp_alpha=ccp_alpha,
                monotonic_cst=monotonic_cst
            )

        super().__init__(max_leaf_nodes=max_leaf_nodes,
                         min_samples_split=min_samples_split,
                         max_depth=max_depth,
                         prune_useless_leaves=prune_useless_leaves,
                         base_stumps=base_stumps,
                         stump_selection=stump_selection,
                         random_state=random_state)

        self.max_depth = max_depth
        self.criterion = criterion
        self.splitter = splitter
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.random_state = random_state
        self.min_impurity_decrease = min_impurity_decrease
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.monotonic_cst = monotonic_cst
        self.distance_matrix = distance_matrix
        self.distance_measure = distance_measure

    def is_split_useless(self, X, clf: tree, idx: np.ndarray):
        """
        Check if a split is useless.

        Args:
            X (np.ndarray): Feature matrix.
            clf (tree): Classifier used for splitting.
            idx (np.ndarray): Indices of the samples to consider.

        Returns:
            bool: True if the split is useless, False otherwise.
        """
        labels = clf.apply(X[idx])

        return len(np.unique(labels)) == 1

    def check_additional_halting_condition(self, y, curr_idx: np.ndarray):
        """
        Check additional halting condition for splitting.

        Args:
            y (np.ndarray): Target labels.
            curr_idx (np.ndarray): Indices of the samples to consider.

        Returns:
            bool: True if halting condition is met, False otherwise.
        """
        return len(np.unique(y[curr_idx])) == 1  # only 1 target

    def queue_push(self, node: RuleTreeNode, idx: np.ndarray):
        """
        Push a node into the priority queue.

        Args:
            node (RuleTreeNode): Node to push.
            idx (np.ndarray): Indices of the samples associated with the node.
        """
        heapq.heappush(self.queue, (len(node.node_id), next(self.tiebreaker), idx, node))

    def make_split(self, X: np.ndarray, y, idx: np.ndarray, medoids_index=None, sample_weight=None, **kwargs) -> tree:
        """
        Create a split using the selected stump.

        Args:
            X (np.ndarray): Feature matrix.
            y (np.ndarray): Target labels.
            idx (np.ndarray): Indices of the samples to consider.
            medoids_index (list): Indices of medoids.
            sample_weight (np.ndarray): Sample weights.
            **kwargs: Additional arguments for splitting.

        Returns:
            tree: The selected stump after fitting.
        """
        pivots_list = ['PivotTreeStumpClassifier',
                       'MultiplePivotTreeStumpClassifier',
                       'ObliquePivotTreeStumpClassifier',
                       'MultipleObliquePivotTreeStumpClassifier']

        if self.stump_selection == 'random':
            stump = self._get_random_stump(X)

            if stump.__class__.__module__.split('.')[-1] in pivots_list:

                stump.fit(X[idx], y[idx], distance_matrix=self.distance_matrix[idx][:, idx], idx=idx,

                          distance_measure=self.distance_measure,
                          sample_weight=None if sample_weight is None else sample_weight[idx])
            else:
                stump.fit(X=X,
                          y=y,
                          idx=idx,
                          context=self,
                          sample_weight=None if sample_weight is None else sample_weight[idx])

        elif self.stump_selection == 'best':
            clfs = []
            info_gains = []
            for _, stump in self._filter_types(X):
                stump = sklearn.clone(stump)

                if stump.__class__.__module__.split('.')[-1] in pivots_list:

                    stump.fit(X=X[idx],
                              y=y[idx],
                              distance_matrix=self.distance_matrix[idx][:, idx],
                              idx=idx,
                              distance_measure=self.distance_measure,
                              sample_weight=None if sample_weight is None else sample_weight[idx])
                else:
                    stump.fit(X=X,
                              y=y,
                              idx=idx,
                              context=self,
                              sample_weight=None if sample_weight is None else sample_weight[idx])

                gain = get_info_gain(stump)
                info_gains.append(gain)

                clfs.append(stump)

            stump = clfs[np.argmax(info_gains)]
        else:
            raise TypeError('Unknown stump selection method')

        return stump

    def prepare_node(self, y: np.ndarray, idx: np.ndarray, node_id: str) -> RuleTreeNode:
        """
        Prepare a node with predictions and probabilities.

        Args:
            y (np.ndarray): Target labels.
            idx (np.ndarray): Indices of the samples to consider.
            node_id (str): Identifier for the node.

        Returns:
            RuleTreeNode: The prepared node.
        """
        prediction = calculate_mode(y[idx])
        predict_proba = np.zeros((len(self.classes_),))
        for i, classe in enumerate(self.classes_):
            predict_proba[i] = sum(np.where(y[idx] == classe, 1, 0)) / len(y[idx])

        return RuleTreeNode(
            node_id=node_id,
            prediction=prediction,
            prediction_probability=predict_proba,
            classes=self.classes_,
            features=self.n_features,
            parent=None,
            stump=None,
            node_l=None,
            node_r=None,
            samples=len(y[idx]),
        )

    def compute_medoids(self, X: np.ndarray, y, idx: np.ndarray, **kwargs):
        """
        Compute medoids for the given data.

        Args:
            X (np.ndarray): Feature matrix.
            y (np.ndarray): Target labels.
            idx (np.ndarray): Indices of the samples to consider.
            **kwargs: Additional arguments for medoid computation.

        Returns:
            list: Indices of the computed medoids.
        """
        if self.distance_measure is not None:
            medoids = []
            sub_matrix = None
            for label in set(y[idx]):
                idx_local_label = np.where(y[idx] == label)[0]
                idx_label = idx[idx_local_label]
                X_class_points = X[idx_label]

                if self.distance_matrix is not None:
                    sub_matrix = self.distance_matrix[idx_label][:, idx_label]
                else:
                    sub_matrix = pairwise_distances(X_class_points, metric=self.distance_measure)
                total_distances = sub_matrix.sum(axis=1)
                medoid_index = idx_label[total_distances.argmin()]
                medoids += [medoid_index]

            return medoids

    def fit(self, X: np.array, y: np.array = None, sample_weight=None, **kwargs):
        """
        Fit the RuleTreeClassifier to the provided data.

        Args:
            X (np.array): Feature matrix.
            y (np.array): Target labels.
            sample_weight (np.array): Sample weights.
            **kwargs: Additional arguments for fitting.

        Returns:
            RuleTreeClassifier: The fitted RuleTreeClassifier instance.
        """
        # Check and initialize the distance matrix if needed
        if self.distance_matrix is None and self.base_stumps is not None:
            base_stumps = self.base_stumps if isinstance(self.base_stumps, list) else [self.base_stumps]
            for stump in base_stumps:
                # Check if the class name matches the specified list
                if stump.__class__.__module__.split('.')[-1] in [
                    'PivotTreeStumpClassifier',
                    'MultiplePivotTreeStumpClassifier',
                    'ObliquePivotTreeStumpClassifier',
                    'MultipleObliquePivotTreeStumpClassifier'
                ]:
                    # Compute the distance matrix
                    self.distance_matrix = pairwise_distances(X, metric=self.distance_measure)
                    # print(X[0][0])
                    # print('compute dist')
                    # print(self.distance_matrix.shape)

                    break  # Distance matrix is initialized, no need to continue

        super().fit(X, y, sample_weight=sample_weight, **kwargs)
        if self.distance_matrix is not None:
            self.distance_matrix = None  # remove to save space when training many estimators

        return self

    def predict_proba(self, X: np.ndarray):
        """
        Predict class probabilities for the input data.

        Args:
            X (np.ndarray): Feature matrix.

        Returns:
            np.ndarray: Predicted class probabilities.
        """
        labels, leaves, proba = self.root.predict(X)

        return proba

    def get_pivots(self, current_node=None, pivot_dicts=None):
        """
        Retrieve pivot information from the tree.

        Args:
            current_node (RuleTreeNode): Current node in the tree.
            pivot_dicts (dict): Dictionary to store pivot information.

        Returns:
            dict: Pivot information for the tree.
        """
        stump_split_map = {
            'PivotTreeStumpClassifier': 'pivot_split',
            'MultiplePivotTreeStumpClassifier': 'multi_pivot_split',
            'ObliquePivotTreeStumpClassifier': 'obl_pivot_split',
            'MultipleObliquePivotTreeStumpClassifier': 'multi_oblique_pivot_split',

            'ObliqueDecisionTreeStumpClassifier': 'oblique_split',
            'DecisionTreeStumpClassifier': 'univariate_split'
        }

        # Initialize pivot_dicts if not provided
        if pivot_dicts is None:
            pivot_dicts = {}

        # Start from root if current_node is not provided
        if current_node is None:
            current_node = self.root

        # Process current node
        if current_node.stump is not None:

            stump_name = current_node.stump.__class__.__module__.split('.')[-1]
            used = current_node.stump.feature_original[0]
            if stump_split_map[stump_name] == 'pivot_split':
                used = [int(used)]
            if stump_split_map[stump_name] == 'multi_pivot_split':
                used = list(used)
            if stump_split_map[stump_name] == 'multi_oblique_pivot_split':
                used = list(used)
            if stump_split_map[stump_name] == 'obl_pivot_split':
                used = [int(x) for x in used]

            if stump_name in stump_split_map:
                if stump_split_map[stump_name] in ['oblique_split', 'univariate_split']:
                    pivot_dicts[current_node.node_id] = {'descriptives': current_node.medoids_index}

                if stump_split_map[stump_name] in ['pivot_split', 'multi_pivot_split', 'multi_oblique_pivot_split',
                                                   'obl_pivot_split']:
                    split_obj = getattr(current_node.stump, stump_split_map[stump_name])
                    if not current_node.is_leaf():
                        pivot_dicts[current_node.node_id] = {
                            'discriminatives': split_obj.get_discriminative_names(),
                            'descriptives': split_obj.get_descriptive_names(),
                            'candidates': split_obj.get_candidates_names(),
                            'used': used
                        }
                    else:
                        pivot_dicts[current_node.node_id] = {'descriptives': split_obj.get_descriptive_names()}



        else:
            if current_node.is_leaf():
                pivot_dicts[current_node.node_id] = {'descriptives': current_node.medoids_index}

        # Recurse into child nodes if they exist
        if current_node.node_l:
            self.get_pivots(current_node.node_l, pivot_dicts)
            self.get_pivots(current_node.node_r, pivot_dicts)

        return pivot_dicts

    def _get_stumps_base_class(self):
        """
        Get the base class for stumps.

        Returns:
            type: The base class for stumps.
        """
        return ClassifierMixin

    def _get_prediction_probas(self, current_node=None, probas=None):
        """
        Retrieve prediction probabilities from the tree.

        Args:
            current_node (RuleTreeNode): Current node in the tree.
            probas (list): List to store probabilities.

        Returns:
            list: Prediction probabilities for the tree.
        """
        if probas is None:
            probas = []

        if current_node is None:
            current_node = self.root

        if current_node.prediction is not None:
            probas.append(current_node.prediction_probability)

        if current_node.node_l:
            self._get_prediction_probas(current_node.node_l, probas)
            self._get_prediction_probas(current_node.node_r, probas)

        return probas

    def local_interpretation(self, X, joint_contribution=False):
        """
        Perform local interpretation of the tree.

        Args:
            X (np.ndarray): Feature matrix.
            joint_contribution (bool): Whether to consider joint contributions.

        Returns:
            np.ndarray: Contributions for each feature and class.
        """
        leaves, paths, leaf_to_path, values = super().local_interpretation(X=X,
                                                                           joint_contribution=joint_contribution)
        normalizer = values.sum(axis=1)[:, np.newaxis]
        normalizer[normalizer == 0.0] = 1.0
        values /= normalizer

        biases = np.tile(values[paths[0][0]], (X.shape[0], 1))
        line_shape = (X.shape[1], self.n_classes_)

        return super().eval_contributions(
            leaves=leaves,
            paths=paths,
            leaf_to_path=leaf_to_path,
            values=values,
            biases=biases,
            line_shape=line_shape,
            joint_contribution=joint_contribution
        )

    def get_balanced_stumps(self, current_node=None, stumps=None, p=0.2):
        """
        Retrieve balanced stumps from the tree.

        Args:
            current_node (RuleTreeNode): Current node in the tree.
            stumps (dict): Dictionary to store balanced stumps.
            p (float): Threshold for balance score.

        Returns:
            dict: Balanced stumps in the tree.
        """
        if stumps is None:
            stumps = {}
        if current_node is None:
            current_node = self.root

        if not current_node.is_leaf():
            if current_node.balance_score > p:
                stumps[current_node.node_id] = (current_node, current_node.balance_score)
                self.get_balanced_stumps(current_node=current_node.node_l, stumps=stumps, p=p)
                self.get_balanced_stumps(current_node=current_node.node_r, stumps=stumps, p=p)

        return stumps

    def stumps_to_trees(self, balanced_nodes):
        """
        Convert balanced stumps to trees.

        Args:
            balanced_nodes (dict): Balanced stumps.

        Returns:
            dict: Trees created from balanced stumps.
        """
        trees = {}

        for k, v in balanced_nodes.items():
            rt = self.__class__()
            rt.classes_ = self.classes_

            node = copy.deepcopy(v[0])
            node_l = copy.deepcopy(node.node_l)
            node_r = copy.deepcopy(node.node_r)

            feat = tuple(node.stump.feature_original[0])  # Ensure it's a tuple

            thr = (node.stump.threshold_original[0],)

            rt.root = node
            rt.root.node_l = node_l
            rt.root_node_r = node_r

            rt.root.node_l.make_leaf()
            rt.root.node_r.make_leaf()

            rt.root.node_id, rt.root.node_l.node_id, rt.root.node_r.node_id = 'R', 'Rl', 'Rr'

            trees[(feat, thr)] = rt

        return trees

    @classmethod
    def complete_tree(cls, node, X, y, n_classes_):
        """
        Complete the tree by filling missing attributes.

        Args:
            node (RuleTreeNode): Current node in the tree.
            X (np.ndarray): Feature matrix.
            y (np.ndarray): Target labels.
            n_classes_ (int): Number of classes.
        """
        classes_ = [i for i in range(n_classes_)]
        node.prediction = calculate_mode(y)
        node.prediction_probability = np.zeros((len(classes_),))
        node.samples = len(y)

        for i, classe in enumerate(classes_):
            node.prediction_probability[i] = np.sum(np.where(y == classe, 1, 0)) / len(y)

        if not node.is_leaf():
            labels_clf = node.stump.apply(X)
            X_l, X_r = X[labels_clf == 1], X[labels_clf == 2]
            y_l, y_r = y[labels_clf == 1], y[labels_clf == 2]
            if X_l.shape[0] != 0:
                cls.complete_tree(node.node_l, X_l, y_l, n_classes_)
            if X_r.shape[0] != 0:
                cls.complete_tree(node.node_r, X_r, y_r, n_classes_)

    @classmethod
    def decode_ruletree(cls, vector, n_features_in_, n_classes_, n_outputs_, numerical_idxs=None,
                        categorical_idxs=None):
        """
        Decode a RuleTree from a vector representation.

        Args:
            vector (list): Vector representation of the tree.
            n_features_in_ (int): Number of input features.
            n_classes_ (int): Number of classes.
            n_outputs_ (int): Number of outputs.
            numerical_idxs (list): Indices of numerical features.
            categorical_idxs (list): Indices of categorical features.

        Returns:
            RuleTreeClassifier: Decoded RuleTreeClassifier instance.
        """
        idx_to_node = super().decode_ruletree(vector)

        for index in range(len(vector[0])):
            # if leaf
            if vector[0][index] == -1:
                idx_to_node[index].prediction = vector[1][index]
            else:
                clf = DecisionTreeStumpClassifier()  ##add kwargs in the function
                clf.numerical = numerical_idxs
                clf.categorical = categorical_idxs
                if isinstance(vector[1][index], str):
                    clf = configure_cat_split(clf, vector[0][index], vector[1][index])
                else:
                    clf = configure_non_cat_split(clf, vector, index,
                                                  n_features_in_, n_classes_, n_outputs_)

                idx_to_node[index].stump = clf
                set_node_children(idx_to_node, index, vector)

        rule_tree = RuleTreeClassifier()
        rule_tree.classes_ = [i for i in range(n_classes_)]
        simplify_decode(idx_to_node[0])
        rule_tree.root = idx_to_node[0]
        return rule_tree

    @classmethod
    def _decode_old(cls, vector, n_features_in_, n_classes_, n_outputs_,
                    numerical_idxs=None, categorical_idxs=None, criterion=None):
        """
        Decode a RuleTree using an older method.

        Args:
            vector (list): Vector representation of the tree.
            n_features_in_ (int): Number of input features.
            n_classes_ (int): Number of classes.
            n_outputs_ (int): Number of outputs.
            numerical_idxs (list): Indices of numerical features.
            categorical_idxs (list): Indices of categorical features.
            criterion (str): Splitting criterion.

        Returns:
            RuleTreeClassifier: Decoded RuleTreeClassifier instance.
        """
        idx_to_node = super().decode_ruletree(vector, n_features_in_, n_classes_, n_outputs_,
                                              numerical_idxs, categorical_idxs, criterion)

        for index in range(len(vector[0])):
            if vector[0][index] == -1:
                idx_to_node[index].prediction = vector[1][index]
            else:
                clf = DecisionTreeStumpClassifier(
                    criterion=criterion)

                clf = DecisionTreeStumpClassifier()

                if numerical_idxs is not None:
                    clf.numerical = numerical_idxs

                if categorical_idxs is not None:
                    clf.categorical = categorical_idxs

                if isinstance(vector[1][index], str):
                    configure_cat_split(clf, vector[0][index], vector[1][index])
                else:
                    configure_non_cat_split(clf, vector, index,
                                            n_features_in_, n_classes_, n_outputs_)
                idx_to_node[index].stump = clf
                set_node_children(idx_to_node, index, vector)

                print(clf)

        rule_tree = RuleTreeClassifier()
        simplify_decode(idx_to_node[0])
        rule_tree.root = idx_to_node[0]
        return rule_tree



from RuleTree.base.RuleTreeBaseStump import RuleTreeBaseStump
from RuleTree.stumps.classification.DecisionTreeStumpClassifier import DecisionTreeStumpClassifier
from RuleTree.stumps.splitters.ObliqueBivariateSplit import ObliqueBivariateSplit
from RuleTree.stumps.splitters.ObliqueHouseHolderSplit import ObliqueHouseHolderSplit
from RuleTree.utils import MODEL_TYPE_CLF


class ObliqueDecisionTreeStumpClassifier(DecisionTreeStumpClassifier, RuleTreeBaseStump):
    """
    A Decision Tree Stump Classifier that uses oblique splits for classification tasks.

    Oblique splits are hyperplanes that are not parallel to the feature axes, allowing
    for more flexible decision boundaries that can capture complex relationships between features.
    This implementation supports both Householder and Bivariate splitting methods.

    Parameters
    ----------
    oblique_split_type : str, default='householder'
        Type of oblique split to use. Options are 'householder' or 'bivariate'.
    pca : object or None, default=None
        PCA instance to use for feature transformation. If None, no PCA is applied.
    max_oblique_features : int, default=2
        Maximum number of features to consider in an oblique split.
    tau : float, default=1e-4
        Threshold parameter for Householder splits.
    n_orientations : int, default=10
        Number of orientations to consider for bivariate splits.
    **kwargs : dict
        Additional parameters to pass to the parent classifier.
    """

    def __init__(self,
                 oblique_split_type='householder',
                 pca=None,
                 max_oblique_features=2,
                 tau=1e-4,
                 n_orientations=10,
                 **kwargs):
        super().__init__(**kwargs)
        self.pca = pca
        self.max_oblique_features = max_oblique_features
        self.tau = tau
        self.n_orientations = n_orientations
        self.oblique_split_type = oblique_split_type

        if self.oblique_split_type == 'householder':
            self.oblique_split = ObliqueHouseHolderSplit(ml_task=MODEL_TYPE_CLF,
                                                         pca=self.pca,
                                                         max_oblique_features=self.max_oblique_features,
                                                         tau=self.tau,
                                                         **kwargs)

        if self.oblique_split_type == 'bivariate':
            self.oblique_split = ObliqueBivariateSplit(ml_task=MODEL_TYPE_CLF, n_orientations=self.n_orientations,
                                                       **kwargs)

    def fit(self, X, y, idx=None, context=None, sample_weight=None, check_input=True):
        """
        Build an oblique decision tree stump classifier from the training set (X, y).

        The method first transforms features using the specified oblique split method and
        then fits a decision tree stump on the transformed data.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training input samples.
        y : array-like of shape (n_samples,)
            The target values (class labels).
        idx : array-like or slice, default=None
            Indices of samples to use for training. If None, all samples are used.
        context : object, default=None
            Additional context information (not used in this implementation).
        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, samples are equally weighted.
        check_input : bool, default=True
            Whether to check input consistency.

        Returns
        -------
        self : object
            Fitted estimator.
        """
        if idx is None:
            idx = slice(None)
        X = X[idx]
        y = y[idx]

        self.feature_analysis(X, y)
        self.num_pre_transformed = self.numerical
        self.cat_pre_transformed = self.categorical
        best_info_gain = -float('inf')

        if len(self.numerical) > 0:
            self.oblique_split.fit(X[:, self.numerical], y, sample_weight=sample_weight, check_input=check_input)
            X_transform = self.oblique_split.transform(X[:, self.numerical])
            super().fit(X_transform, y, sample_weight=sample_weight, check_input=check_input)

            self.feature_original = [self.oblique_split.feats, -2, -2]
            self.coefficients = self.oblique_split.coeff
            self.threshold_original = self.tree_.threshold
            self.is_oblique = True

        return self

    def apply(self, X):
        """
        Apply the decision tree stump to X, returning leaf indices.

        This method transforms the input features using the fitted oblique split method
        and then applies the decision tree stump to the transformed features.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        X_leaves : array-like of shape (n_samples,)
            For each datapoint x in X, return the index of the leaf x ends up in.
        """
        X_transform = self.oblique_split.transform(X[:, self.num_pre_transformed])
        return super().apply_sk(X_transform)  # otherwise we need to "personalize" the apply function

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.

        Parameters
        ----------
        deep : bool, default=True
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        return {
            **self.kwargs,
            'oblique_split_type': self.oblique_split_type,
            'max_oblique_features': self.max_oblique_features,
            'pca': self.pca,
            'tau': self.tau,
            'n_orientations': self.n_orientations
        }

    def get_rule(self, columns_names=None, scaler=None, float_precision=3):
        """
        Extract the decision rule from the stump in a human-readable format.

        This method creates different representations of the oblique decision rule
        (textual, blob, graphviz) for visualization and interpretation purposes.

        Parameters
        ----------
        columns_names : array-like, default=None
            Names of the features. If None, feature indices are used.
        scaler : object, default=None
            Scaler object used to transform features. If provided, the rule is
            expressed in terms of the original (unscaled) feature values.
        float_precision : int, default=3
            Number of decimal places to use for floating point values in the rule.
            If None, no rounding is performed.

        Returns
        -------
        rule : dict
            Dictionary containing different representations of the rule:
            - feature_idx: indices of features used in the rule
            - threshold: threshold value for the rule
            - coefficients: coefficients of the oblique split
            - is_categorical: whether the split is on categorical features
            - samples: number of samples covered by the rule
            - feature_name: string representation of the features
            - textual_rule: human-readable rule representation
            - blob_rule: compact rule representation
            - graphviz_rule: rule representation for graphviz visualization
            - not_textual_rule: negation of the rule
            - not_blob_rule: negation of the compact rule
            - not_graphviz_rule: negation of the rule for graphviz
        """
        rule = {
            "feature_idx": self.feature_original[0],  # list of feats
            "threshold": self.threshold_original[0],  # thr
            "coefficients": self.coefficients,  # coefficients
            "is_categorical": self.is_categorical,
            "samples": self.n_node_samples[0]
        }

        # round coefficients here
        rule['coefficients'] = [
            str(coeff) if float_precision is None else round(float(coeff), float_precision)
            for coeff in rule['coefficients']
        ]

        feat_name = " + ".join(f"{coeff} * X_{idx}" for coeff, idx in zip(rule['coefficients'], rule['feature_idx']))

        if columns_names is not None:
            feat_name = "_".join(columns_names[idx] for idx in self.feature_original[0])  # check this for feat names
        rule["feature_name"] = feat_name

        if scaler is not None:
            # TODO
            raise NotImplementedError()
            pass

        comparison = "<=" if not self.is_categorical else "="
        not_comparison = ">" if not self.is_categorical else "!="
        rounded_value = str(rule["threshold"]) if float_precision is None else round(rule["threshold"], float_precision)

        if scaler is not None:
            # TODO
            raise NotImplementedError()
            pass

        rule["textual_rule"] = f"{feat_name} {comparison} {rounded_value}\t{rule['samples']}"
        rule["blob_rule"] = f"{feat_name} {comparison} {rounded_value}"
        rule["graphviz_rule"] = {
            "label": f"{feat_name} {'\u2264' if not self.is_categorical else '='} {rounded_value}",
        }

        rule["not_textual_rule"] = f"{feat_name} {not_comparison} {rounded_value}"
        rule["not_blob_rule"] = f"{feat_name} {not_comparison} {rounded_value}"
        rule["not_graphviz_rule"] = {
            "label": f"{feat_name} {'>' if not self.is_categorical else '\u2260'} {rounded_value}"
        }

        return rule


def node_to_dict(self):
    """
    Convert the decision stump node to a dictionary representation.

    This method is useful for serialization and visualization of the model.

    Returns
    -------
    rule : dict
        Dictionary containing the rule information and additional metadata:
        - stump_type: class name of the stump
        - samples: number of samples covered by the stump
        - impurity: impurity measure at the stump
        - args: dictionary of additional arguments including oblique-specific parameters
        - split: dictionary containing split arguments
    """
    rule = self.get_rule(float_precision=None)

    rule["stump_type"] = self.__class__.__name__
    rule["samples"] = self.n_node_samples[0]
    rule["impurity"] = self.tree_.impurity[0]

    rule["args"] = {
                       "is_oblique": self.is_oblique,
                       "is_pivotal": self.is_pivotal,
                       "unique_val_enum": self.unique_val_enum,
                       "coefficients": self.coefficients,
                   } | self.kwargs

    rule["split"] = {
        "args": {}
    }

    return rule



import time
import random

import numba
import numpy as np
import pandas as pd
import scipy.io.arff

from numba import jit, prange, UnsupportedError
import psutil
from sklearn.base import TransformerMixin
from sklearn.ensemble import RandomForestClassifier
from sklearn.feature_selection import mutual_info_classif, mutual_info_regression
from sklearn.metrics import classification_report
from sklearn.utils import resample

from RuleTree.utils.shapelet_transform.matrix_to_vector_distances import euclidean, sqeuclidean, cosine, cityblock


class Shapelets(TransformerMixin):
    """
    Shapelets transformer for time series classification.
    
    This class implements the shapelet transform method which transforms time series data
    into a feature vector of distances between the time series and a set of shapelets.
    Shapelets are discriminative subsequences extracted from the time series.
    
    The transformer follows scikit-learn's TransformerMixin interface (fit, transform).
    
    Parameters
    ----------
    n_shapelets : int, default=100
        Number of shapelets to be used in the transformation.
    
    n_shapelets_for_selection : int or str, default=np.inf
        Number of candidate shapelets to generate before selection.
        If 'stratified', generates candidates based on class distribution.
        If np.inf, generates all possible candidates.
    
    n_ts_for_selection_per_class : int, default=np.inf
        Number of time series to use per class for shapelet candidates.
    
    sliding_window : int, default=50
        Length of the shapelets to extract.
    
    selection : str, default='random'
        Method used to select the final shapelets from candidates:
        - 'random': random selection
        - 'mi_clf': mutual information (classification)
        - 'mi_reg': mutual information (regression)
        - 'cluster': KMedoids clustering
    
    distance : str or callable, default='euclidean'
        Distance metric to use for computing distances between shapelets and time series.
        Options: 'euclidean', 'sqeuclidean', 'cosine', 'cityblock', or custom callable.
    
    mi_n_neighbors : int, default=100
        Number of neighbors to use for mutual information calculation.
    
    random_state : int, default=42
        Random seed for reproducibility.
    
    n_jobs : int, default=1
        Number of parallel jobs. If -1, uses all available processors.
    """
    __distances = {
        'euclidean': euclidean,
        'sqeuclidean': sqeuclidean,
        'cosine': cosine,
        'cityblock': cityblock,
    }
    def __init__(self,
                 n_shapelets=100,
                 n_shapelets_for_selection=np.inf, #int, inf, or 'stratified'
                 n_ts_for_selection_per_class=np.inf, #int, inf
                 sliding_window=50,
                 selection='random', #random, mi_clf, mi_reg, cluster
                 distance='euclidean',
                 mi_n_neighbors = 100,
                 random_state=42, n_jobs=1):
        super().__init__()

        self.shapelets = None
        if n_jobs == -1:
            n_jobs = psutil.cpu_count()

        if isinstance(distance, str) and distance not in self.__distances:
            raise UnsupportedError(f"Unsupported distance '{distance}'")

        if selection not in ["random", "mi_clf", "mi_reg", "cluster"]:
            raise UnsupportedError(f"Unsupported selection '{selection}'")

        self.n_shapelets = n_shapelets
        self.n_shapelets_for_selection = n_shapelets_for_selection
        self.n_ts_for_selection_per_class = n_ts_for_selection_per_class
        self.sliding_window = sliding_window
        self.selection = selection
        self.distance = distance
        self.mi_n_neighbors = mi_n_neighbors
        self.random_state = random_state
        self.n_jobs = n_jobs

        random.seed(random_state)

    def __get_distance(self):
        """
        Get the distance function based on the distance parameter.
        
        Returns
        -------
        callable
            The distance function to use for computing shapelet distances.
        """
        if isinstance(self.distance, str):
            return self.__distances[self.distance]
        return self.distance


    def fit(self, X, y=None, **fit_params):
        """
        Fit the Shapelets transformer on training data.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_signals, n_timepoints)
            Training time series data.
            Currently only supports n_signals=1 (univariate time series).
        
        y : array-like of shape (n_samples,), default=None
            Target values for supervised learning.
            
        **fit_params : dict
            Additional parameters passed to selection methods.
            
        Returns
        -------
        self : Shapelets
            Returns self.
            
        Raises
        ------
        NotImplementedError
            If multivariate time series are provided (n_signals > 1).
        """
        # X.shape = (n_records, n_signals, n_obs)
        if X.shape[1] != 1:
            raise NotImplementedError("Multivariate TS are not supported (yet).")

        candidate_shapelets = self.__fit_partition(X, y)

        if self.selection == 'random':
            self.shapelets = self.__fit_selection_random(candidate_shapelets, X, y)
        elif self.selection == 'mi_clf':
            self.shapelets = self.__fit_selection_mutual_info(candidate_shapelets, X, y, mutual_info_classif)
        elif self.selection == 'mi_reg':
            self.shapelets = self.__fit_selection_mutual_info(candidate_shapelets, X, y, mutual_info_regression)
        elif self.selection == 'cluster':
            self.shapelets = self.__fit_selection_cluster(candidate_shapelets, X, y)

        return self




    def __fit_partition(self, X, y):
        """
        Generate candidate shapelets from the time series data.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_signals, n_timepoints)
            Time series data.
            
        y : array-like of shape (n_samples,), default=None
            Target values.
            
        Returns
        -------
        array-like of shape (n_candidates, n_signals, shapelet_length)
            Candidate shapelets extracted from the time series data.
        """
        if y is None:
            y = np.zeros(X.shape[0])

        classes = np.unique(y)
        n_classes = len(classes)

        candidate_shapelets = []

        if not isinstance(self.n_shapelets_for_selection, str) and np.isinf(self.n_shapelets_for_selection):
            for ts_idx in range(X.shape[0]):
                for position_idx in range(X.shape[-1] - self.sliding_window):
                    candidate_shapelets.append(X[ts_idx, :, position_idx: position_idx + self.sliding_window])

            return np.array(candidate_shapelets)

        if self.n_shapelets_for_selection == 'stratified':
            classes, n_candidate_per_class = np.unique(
                resample(y, stratify=y, random_state=self.random_state, n_samples=int(len(y)**.5)), return_counts=True)
        n_candidate_per_class = [max(1, round(self.n_shapelets_for_selection/n_classes)) for _ in classes]


        for classe in classes:
            X_class = X[y == classe]
            for _ in n_candidate_per_class:
                ts_idx = random.randint(0, len(X_class) - 1)
                start = random.randint(0, X_class.shape[-1] - self.sliding_window)
                stop = start + self.sliding_window
                candidate_shapelets.append(np.copy(X_class[ts_idx, :, start:stop]))

        return np.array(candidate_shapelets)

    def __fit_selection_random(self, candidate_shapelets: np.ndarray, X, y):
        """
        Randomly select shapelets from candidates.
        
        Parameters
        ----------
        candidate_shapelets : array-like of shape (n_candidates, n_signals, shapelet_length)
            Candidate shapelets.
        
        X : array-like of shape (n_samples, n_signals, n_timepoints)
            Time series data (unused in random selection).
            
        y : array-like of shape (n_samples,)
            Target values (unused in random selection).
            
        Returns
        -------
        array-like of shape (n_shapelets, n_signals, shapelet_length)
            Randomly selected shapelets.
        """
        n_shapelets = min(self.n_shapelets, candidate_shapelets.shape[0])
        return candidate_shapelets[np.random.choice(candidate_shapelets.shape[0], size=n_shapelets, replace=False)]

    def __fit_selection_mutual_info(self, candidate_shapelets: np.ndarray, X, y, mutual_info_fun):
        """
        Select shapelets using mutual information criteria.
        
        Parameters
        ----------
        candidate_shapelets : array-like of shape (n_candidates, n_signals, shapelet_length)
            Candidate shapelets.
            
        X : array-like of shape (n_samples, n_signals, n_timepoints)
            Time series data.
            
        y : array-like of shape (n_samples,)
            Target values.
            
        mutual_info_fun : callable
            Function to compute mutual information (either mutual_info_classif or mutual_info_regression).
            
        Returns
        -------
        array-like of shape (n_shapelets, n_signals, shapelet_length)
            Shapelets selected based on mutual information.
            
        Raises
        ------
        UnsupportedError
            If y is None (unsupervised tasks).
        """
        if y is None:
            raise UnsupportedError("Mutual information is not suitable for unsupervised tasks.")

        idx_to_test = resample(range(X.shape[0]), stratify=y, random_state=self.random_state)

        old_n_threads = numba.get_num_threads()
        numba.set_num_threads(self.n_jobs)
        dist = _best_fit(X[idx_to_test], candidate_shapelets, self.__get_distance())
        numba.set_num_threads(old_n_threads)

        labels, labels_count = np.unique(y, return_counts=True)
        if np.sum(labels_count) == len(labels):
            scores = np.zeros((len(labels), ))
        else:
            scores = mutual_info_fun(dist, y,
                                     n_jobs=self.n_jobs,
                                     n_neighbors=min(dist.shape[0], self.mi_n_neighbors),
                                     discrete_features=False)
        if len(candidate_shapelets) == self.n_shapelets:
            return candidate_shapelets
        return candidate_shapelets[np.argpartition(scores, -min(scores.shape[0], self.n_shapelets))\
            [-min(scores.shape[0], self.n_shapelets):]]

    def __fit_selection_cluster(self, candidate_shapelets, X, y):
        """
        Select shapelets using KMedoids clustering.
        
        Parameters
        ----------
        candidate_shapelets : array-like of shape (n_candidates, n_signals, shapelet_length)
            Candidate shapelets.
            
        X : array-like of shape (n_samples, n_signals, n_timepoints)
            Time series data (unused in cluster selection).
            
        y : array-like of shape (n_samples,)
            Target values (unused in cluster selection).
            
        Returns
        -------
        array-like of shape (n_shapelets, n_signals, shapelet_length)
            Shapelets selected as medoids from clustering.
            
        Raises
        ------
        Exception
            If scikit-learn-extra is not installed.
        """
        old_n_threads = numba.get_num_threads()
        numba.set_num_threads(self.n_jobs)
        dist_matrix = _best_fit(candidate_shapelets, candidate_shapelets, self.__get_distance())
        numba.set_num_threads(old_n_threads)

        try:
            from sklearn_extra.cluster import KMedoids
            clu = KMedoids(n_clusters=self.n_shapelets, random_state=self.random_state, metric='precomputed')
        except Exception as e:
            raise Exception(f"Please install scikit-learn-extra [{e}]")
        clu.fit(dist_matrix)

        return candidate_shapelets[clu.medoid_indices_]

    def transform(self, X, y=None, **transform_params):
        """
        Transform time series data into shapelet distance features.
        
        For each time series in X, computes the minimum distance to each shapelet,
        resulting in a feature vector of length n_shapelets.
        
        Parameters
        ----------
        X : array-like of shape (n_samples, n_signals, n_timepoints)
            Time series data to transform.
            
        y : array-like of shape (n_samples,), default=None
            Target values (unused in transformation).
            
        **transform_params : dict
            Additional parameters (unused).
            
        Returns
        -------
        array-like of shape (n_samples, n_shapelets)
            Transformed data, where each feature is the minimum distance between
            a time series and a shapelet.
        """
        old_n_threads = numba.get_num_threads()
        numba.set_num_threads(self.n_jobs)
        dist_matrix = _best_fit(X, self.shapelets, self.__get_distance())
        numba.set_num_threads(old_n_threads)

        return dist_matrix

@jit(parallel=True)
def _best_fit(timeseries: np.ndarray, shapelets: np.ndarray, distance):
    """
    Compute the minimum distance between time series and shapelets.
    
    For each time series and shapelet, computes the minimum distance between
    the shapelet and all subsequences of the time series.
    
    Parameters
    ----------
    timeseries : array-like of shape (n_samples, n_signals, n_timepoints)
        Time series data.
        
    shapelets : array-like of shape (n_shapelets, n_signals, shapelet_length)
        Shapelets.
        
    distance : callable
        Distance function to use.
        
    Returns
    -------
    array-like of shape (n_samples, n_shapelets)
        Matrix of minimum distances between time series and shapelets.
    """
    res = np.ones((timeseries.shape[0], shapelets.shape[0]), dtype=np.float32)*np.inf
    w = shapelets.shape[-1]

    for ts_idx in prange(timeseries.shape[0]):
        ts = timeseries[ts_idx, 0, :]
        ts_sw = np.lib.stride_tricks.sliding_window_view(ts, w)
        for shapelet_idx, shapelet in enumerate(shapelets[:, 0, :]):
            distance_matrix = distance(ts_sw, shapelet)
            res[ts_idx, shapelet_idx] = np.min(distance_matrix)

    return res





if __name__ == '__main__':
    df_train = pd.DataFrame(scipy.io.arff.loadarff('test_dataset/CBF/CBF_TRAIN.arff')[0])
    df_test = pd.DataFrame(scipy.io.arff.loadarff('test_dataset/CBF/CBF_TEST.arff')[0])
    df_train.target = df_train.target.astype(int)
    df_test.target = df_test.target.astype(int)

    X_train = df_train.drop(columns=['target']).values.reshape((-1, 1, 128))
    X_test = df_test.drop(columns=['target']).values.reshape((-1, 1, 128))
    y_train = df_train['target'].values
    y_test = df_test['target'].values

    st = Shapelets(n_shapelets=10, selection='random', mi_n_neighbors=100, n_jobs=-1, distance='cityblock') #euclidean, sqeuclidean, cosine, cityblock

    X_train_transform = st.fit_transform(X_train, y_train)
    X_test_transform = st.transform(X_test)

    rf = RandomForestClassifier(n_estimators=100, n_jobs=-1)
    rf.fit(X_train_transform, y_train)

    y_pred = rf.predict(X_test_transform)

    print(classification_report(y_test, y_pred))




import numpy as np
from scipy.stats import wasserstein_distance
from sklearn.preprocessing import KBinsDiscretizer


def _earth_mover_distance(a:np.ndarray, b_idx:np.ndarray, is_categorical:bool):
    """
    Calculate the Earth Mover's Distance between two distributions.
    
    The Earth Mover's Distance (also known as Wasserstein distance) measures
    the minimum "work" required to transform one distribution into another.
    
    Parameters
    ----------
    a : np.ndarray
        Array containing all data points
    b_idx : np.ndarray
        Boolean mask indicating which points in 'a' belong to second distribution
    is_categorical : bool
        Whether the data is categorical or continuous
        
    Returns
    -------
    float
        The Earth Mover's Distance between distributions
    """
    b = a[b_idx]
    len_a = len(a)
    len_b = len(b)

    if is_categorical:
        unique_values, a_binned = np.unique(a, return_counts=True)
        a_binned = a_binned.astype(np.float64)
        a_binned /= len_a
        b_binned = np.asarray([np.sum(b == x) for x in unique_values])/len_b
    else:
        bins = KBinsDiscretizer(n_bins=max(3, int(len_a**.5)), strategy='uniform')
        a_binned = np.sum(bins.fit_transform(a.reshape(-1, 1)), axis=0)[0]/len_a
        b_binned = np.sum(bins.transform(b.reshape(-1, 1)), axis=0)[0]/len_b

    bin_list = np.arange(0, a_binned.shape[0], 1)
    d=wasserstein_distance(bin_list, bin_list, a_binned, b_binned)# / (len_a-1)
    return d

def compute_k_anonimity(X, X_bool, sensible_attribute):
    """
    Compute k-anonymity values for left and right partitions of data.
    
    K-anonymity ensures that each group contains at least k individuals with the same 
    sensitive attribute value, making it difficult to identify a specific individual.
    
    Parameters
    ----------
    X : np.ndarray
        Feature matrix
    X_bool : np.ndarray
        Boolean array indicating the split (True for left partition, False for right)
    sensible_attribute : int
        Index of the sensitive attribute column
        
    Returns
    -------
    tuple
        (k_left, k_right) representing the k-anonymity values for both partitions
    """
    X_bool = X_bool.copy().reshape(-1)
    k_left = min(np.unique(X[X_bool, sensible_attribute], return_counts=True)[1])
    k_right = min(np.unique(X[~X_bool, sensible_attribute], return_counts=True)[1])

    return k_left, k_right

def compute_l_diversity(X, X_bool, sensible_attribute):
    """
    Compute l-diversity values for left and right partitions of data.
    
    L-diversity ensures that sensitive attribute values are well-represented in each group,
    making it difficult to infer an individual's sensitive attribute.
    
    Parameters
    ----------
    X : np.ndarray
        Feature matrix
    X_bool : np.ndarray
        Boolean array indicating the split (True for left partition, False for right)
    sensible_attribute : int
        Index of the sensitive attribute column
        
    Returns
    -------
    tuple
        (l_left, l_right) representing the l-diversity values for both partitions
    """
    X_bool = X_bool.copy().reshape(-1)
    l_left = len(np.unique(X[X_bool, sensible_attribute])) - 1
    l_right = len(np.unique(X[~X_bool, sensible_attribute])) - 1

    return l_left, l_right

def _compute_t_closeness(X, sensible_attribute, categorical):
    """
    Compute t-closeness for a given dataset partition.
    
    T-closeness measures the distance between the distribution of a sensitive attribute 
    in a group and the distribution of the attribute in the entire dataset.
    
    Parameters
    ----------
    X : np.ndarray
        Feature matrix for a partition
    sensible_attribute : int
        Index of the sensitive attribute column
    categorical : list or set
        Indices of categorical features
        
    Returns
    -------
    float
        The minimum t-closeness value across all features
    """
    t = np.inf
    for prot_attr_value in np.unique(X[:, sensible_attribute]):
        for feat_idx in range(X.shape[1]):
            if feat_idx == sensible_attribute:
                continue

            t_curr = _earth_mover_distance(X[:, feat_idx],
                                           X[:, sensible_attribute] == prot_attr_value,
                                           feat_idx in categorical)
            if t_curr < t:
                t = t_curr

    return t

def compute_t_closeness(X, X_bool, sensible_attribute, categorical):
    """
    Compute t-closeness values for left and right partitions of data.
    
    Parameters
    ----------
    X : np.ndarray
        Feature matrix
    X_bool : np.ndarray
        Boolean array indicating the split (True for left partition, False for right)
    sensible_attribute : int
        Index of the sensitive attribute column
    categorical : list or set
        Indices of categorical features
        
    Returns
    -------
    tuple
        (t_left, t_right) representing the t-closeness values for both partitions
    """
    X_bool = X_bool.copy().reshape(-1)

    t_left = _compute_t_closeness(X[X_bool], sensible_attribute, categorical)
    t_right = _compute_t_closeness(X[~X_bool], sensible_attribute, categorical)

    return t_left, t_right



"""
Utility functions for tree decoding and manipulation in RuleTree.

This module provides helper functions to configure, create, and manipulate decision trees,
primarily focusing on split configurations and tree structure modifications.
"""

import numpy as np
from sklearn.tree._tree import Tree


def configure_cat_split(clf, feature_index, threshold_value):
    """
    Configure a categorical split in a classifier.
    
    Parameters
    ----------
    clf : object
        Classifier object to be modified.
    feature_index : int
        Index of the feature used for the categorical split.
    threshold_value : float
        Threshold value for the categorical split.
        
    Returns
    -------
    clf : object
        Modified classifier with the categorical split configuration.
    """
    clf.is_categorical = True
    clf.feature_original = [feature_index - 1, -2, -2]
    clf.threshold_original = np.array([threshold_value, -2, -2])
    return clf
    
def configure_non_cat_split(clf, vector, index, n_features_in_, n_classes_, n_outputs_):
    """
    Configure a non-categorical split in a classifier.
    
    Parameters
    ----------
    clf : object
        Classifier object to be modified.
    vector : list of arrays
        Contains feature indices and threshold values.
    index : int
        Index to determine which split to use from the vector.
    n_features_in_ : int
        Number of input features.
    n_classes_ : int
        Number of classes.
    n_outputs_ : int
        Number of outputs.
        
    Returns
    -------
    clf : object
        Modified classifier with the non-categorical split configuration.
    """
    inner_tree = create_inner_tree(n_features_in_, n_classes_, n_outputs_, vector, index)
    assign_tree_properties(clf, inner_tree)
    return clf


def create_inner_tree(n_features_in_, n_classes_, n_outputs_, vector, index):
    """
    Create an inner tree structure with a specified split.
    
    Parameters
    ----------
    n_features_in_ : int
        Number of input features.
    n_classes_ : int
        Number of classes.
    n_outputs_ : int
        Number of outputs.
    vector : list of arrays
        Contains feature indices and threshold values.
    index : int
        Index to determine which split to use from the vector.
        
    Returns
    -------
    inner_tree : sklearn.tree._tree.Tree
        A tree object with the configured split.
        
    Notes
    -----
    Creates a basic tree structure with a depth of 1 (root and two leaf nodes).
    """
    n_classes_ = np.array([n_classes_], dtype=np.intp)
    inner_tree = Tree(n_features_in_, n_classes_, n_outputs_)
    state = inner_tree.__getstate__()
    
    node_count = 3  # Assuming max_depth 1 implies 3 nodes
    state['node_count'] = node_count
    state['nodes'] = np.zeros((node_count,), 
                              dtype=[('left_child', '<i8'),
                                     ('right_child', '<i8'), 
                                     ('feature', '<i8'), 
                                     ('threshold', '<f8'), 
                                     ('impurity', '<f8'), 
                                     ('n_node_samples', '<i8'), 
                                     ('weighted_n_node_samples', '<f8'),
                                     ('missing_go_to_left', 'u1')])
    state['values'] = np.zeros((node_count, n_outputs_, n_classes_[0]))
    
    state['nodes'][0] = (1, 2, vector[0][index] - 1, vector[1][index], 0.0, 1, 1.0, 0)  # root node with two children
    state['nodes'][1] = (-1, -1, -2, -2.0, 0.0, 1, 1.0, 0)  
    state['nodes'][2] = (-1, -1, -2, -2.0, 0.0, 1, 1.0, 0)  
    inner_tree.__setstate__(state)
    
    return inner_tree

def assign_tree_properties(clf, inner_tree):
    """
    Assign tree properties from an inner tree to a classifier.
    
    Parameters
    ----------
    clf : object
        Classifier object to be modified.
    inner_tree : sklearn.tree._tree.Tree
        The tree structure to assign to the classifier.
        
    Notes
    -----
    This function copies essential properties from the inner_tree to the classifier,
    ensuring that the classifier has the correct tree structure and associated metadata.
    """
    clf.tree_ = inner_tree
    clf.n_outputs_ = inner_tree.n_outputs
    clf.n_classes_ = inner_tree.n_classes[0]
    clf.classes_ = np.arange(inner_tree.n_classes[0])
    clf.n_features_in_ = inner_tree.n_features
    clf.max_features_ = inner_tree.n_features
    clf.feature_original = clf.tree_.feature
    clf.threshold_original = clf.tree_.threshold
    
def set_node_children(idx_to_node, index, vector):
    """
    Set parent-child relationships between nodes in a tree.
    
    Parameters
    ----------
    idx_to_node : dict
        Dictionary mapping indices to node objects.
    index : int
        Index of the parent node.
    vector : list of arrays
        Contains information about the tree structure.
        
    Notes
    -----
    This function establishes the hierarchical relationships between nodes,
    setting up parent-child connections and node IDs for tree traversal.
    """
    left_child_idx = 2 * index + 1
    right_child_idx = 2 * index + 2
    
    if left_child_idx < len(vector[0]):
        idx_to_node[left_child_idx].parent = idx_to_node[index]
        idx_to_node[left_child_idx].node_id = idx_to_node[index].node_id + 'l'

    if right_child_idx < len(vector[0]):
        idx_to_node[right_child_idx].parent = idx_to_node[index]
        idx_to_node[right_child_idx].node_id = idx_to_node[index].node_id + 'r'

    idx_to_node[index].node_l = idx_to_node[left_child_idx]
    idx_to_node[index].node_r = idx_to_node[right_child_idx]


def simplify_decode(node):
    """
    Simplify a tree by removing redundant splits.
    
    This function recursively analyzes a decision tree and simplifies its structure
    by converting internal nodes to leaf nodes when all descendant leaves have the same prediction.
    
    Parameters
    ----------
    node : object
        A node in the decision tree with node_l and node_r attributes referring to child nodes.
        
    Returns
    -------
    set
        A set of unique predictions in the subtree rooted at this node.
        
    Notes
    -----
    If all leaf nodes in a subtree have the same prediction, the function converts
    the root of that subtree to a leaf node with that prediction.
    """
    if node.is_leaf():
        return {node.prediction}
    else:
        all_pred = simplify_decode(node.node_l) | simplify_decode(node.node_r)
        if len(all_pred) == 1:
            node.prediction = node.node_l.prediction
            node.make_leaf()
            return {node.prediction}
        else:
            return all_pred
